"""TcEx Framework Playbook module"""
# standard library
import base64
import json
import re
from collections import OrderedDict


class PlaybooksBase:
    """TcEx Playbook Module Base Class

    Args:
        tcex (TcEx): Instance of TcEx class.
        context (str): The Redis context (hash).
        output_variables (list): The requested output variables.
    """

    def __init__(self, tcex, context, output_variables):
        """Initialize the Class properties."""
        self.tcex = tcex
        self._context = context
        self._output_variables = output_variables or []

        # properties
        self._key_value_store = None
        self._output_variables_by_name = None
        self._output_variables_by_type = None
        self.log = tcex.log

        # match full variable
        self._variable_match = re.compile(fr'^{self._variable_pattern}$')
        # capture variable parts (exactly a variable)
        self._variable_parse = re.compile(self._variable_pattern)
        # match embedded variables without quotes (#App:7979:variable_name!StringArray)
        self._vars_keyvalue_embedded = re.compile(fr'(?:\"\:\s?)[^\"]?{self._variable_pattern}')

    def _coerce_string_value(self, value):
        """Return a string value from an bool or int."""
        # coerce bool before int as python says a bool is an int
        if isinstance(value, bool):
            # coerce bool to str type
            self.log.warning(f'Coercing bool value ({value}) to a string ("{str(value).lower()}").')
            value = str(value).lower()

        # coerce int to str type
        if isinstance(value, (float, int)):
            self.log.warning(f'Coercing float/int value ({value}) to a string ("{str(value)}").')
            value = str(value)

        return value

    def _create(self, key, value, validate=True):
        """Create the value in Redis if applicable."""
        if key is None or value is None:
            self.log.warning('The key or value field is None.')
            return None

        # get variable type from variable value
        variable_type = self.variable_type(key)

        if variable_type == 'Binary':
            # if not isinstance(value, bytes):
            #     value = value.encode('utf-8')
            if validate and not isinstance(value, bytes):
                raise RuntimeError('Invalid data provided for Binary.')
            value = base64.b64encode(value).decode('utf-8')
        elif variable_type == 'KeyValue':
            if validate and (not isinstance(value, dict) or not self._is_key_value(value)):
                raise RuntimeError('Invalid data provided for KeyValue.')
        elif variable_type == 'String':
            # coerce string values
            value = self._coerce_string_value(value)

            if validate and not isinstance(value, str):
                raise RuntimeError('Invalid data provided for String.')
        elif variable_type == 'TCEntity':
            if validate and (not isinstance(value, dict) or not self._is_tc_entity(value)):
                raise RuntimeError('Invalid data provided for TcEntity.')

        # self.log.trace(f'pb create - context: {self._context}, key: {key}, value: {value}')
        try:
            value = json.dumps(value)
        except ValueError as e:  # pragma: no cover
            raise RuntimeError(f'Failed to serialize value ({e}).')

        return self.key_value_store.create(key.strip(), value)

    def _create_array(self, key, value, validate=True):
        """Create the value in Redis if applicable."""
        if key is None or value is None:
            self.log.warning('The key or value field is None.')
            return None

        # get variable type from variable value
        variable_type = self.variable_type(key)

        if variable_type == 'BinaryArray':
            if validate and not isinstance(value, list):
                raise RuntimeError('Invalid data provided for BinaryArray.')

            value_encoded = []
            for v in value:
                if v is not None:
                    if validate and not isinstance(v, bytes):
                        raise RuntimeError('Invalid data provided for Binary.')
                    # if not isinstance(v, bytes):
                    #     v = v.encode('utf-8')
                    v = base64.b64encode(v).decode('utf-8')
                value_encoded.append(v)
            value = value_encoded
        elif variable_type == 'KeyValueArray':
            if validate and (not isinstance(value, list) or not self._is_key_value_array(value)):
                raise RuntimeError('Invalid data provided for KeyValueArray.')
        elif variable_type == 'StringArray':
            if validate and not isinstance(value, list):
                raise RuntimeError('Invalid data provided for StringArray.')

            value_coerced = []
            for v in value:
                # coerce string values
                v = self._coerce_string_value(v)

                if validate and not isinstance(v, (type(None), str)):
                    raise RuntimeError('Invalid data provided for StringArray.')
                value_coerced.append(v)
            value = value_coerced
        elif variable_type == 'TCEntityArray':
            if validate and (not isinstance(value, list) or not self._is_tc_entity_array(value)):
                raise RuntimeError('Invalid data provided for TcEntityArray.')

        # self.log.trace(f'pb create - context: {self._context}, key: {key}, value: {value}')
        try:
            value = json.dumps(value)
        except ValueError as e:  # pragma: no cover
            raise RuntimeError(f'Failed to serialize value ({e}).')

        return self.key_value_store.create(key.strip(), value)

    @staticmethod
    def _decode_binary(data):
        """Return decoded bytes data handling data written by java apps."""
        try:
            data = data.decode('utf-8')
        except UnicodeDecodeError:  # pragma: no cover
            # for data written an upstream java App
            data = data.decode('latin-1')
        return data

    @staticmethod
    def _is_key_value(data):
        """Return True if provided data has proper structure for Key Value."""
        if data is None:
            return False
        return all(x in data for x in ['key', 'value'])

    def _is_key_value_array(self, data):
        """Return True if provided data has proper structure for Key Value Array."""
        for d in data:
            if not self._is_key_value(d):
                return False
        return True

    @staticmethod
    def _is_tc_entity(data):
        """Return True if provided data has proper structure for TC Entity."""
        if data is None:
            return False
        return all(x in data for x in ['id', 'value', 'type'])

    def _is_tc_entity_array(self, data):
        """Return True if provided data has proper structure for TC Entity Array."""
        for d in data:
            if not self._is_tc_entity(d):
                return False
        return True

    @staticmethod
    def _load_value(value):
        """Return the loaded JSON value or raise an error.

        Args:
            value (str): The data from key/value store.

        Raises:
            RuntimeError: Raise error when data can't be loaded as JSON data.

        Returns:
            any: The de-serialized value from the key/value store.
        """
        try:
            return json.loads(value, object_pairs_hook=OrderedDict)
        except ValueError as e:  # pragma: no cover
            raise RuntimeError(f'Failed to JSON load data "{value}" ({e}).')

    def _parse_output_variables(self):
        """Parse the output variables provided to Playbook Class.

        **Example Variable Format**::

            ['#App:1234:status!String', '#App:1234:status_code!String']
        """
        self._output_variables_by_name = {}
        self._output_variables_by_type = {}
        for ov in self._output_variables:
            # parse the variable to get individual parts
            parsed_variable = self.parse_variable(ov)
            variable_name = parsed_variable.get('name')
            variable_type = parsed_variable.get('type')

            # store the variables in dict by name (e.g. "status_code")
            self._output_variables_by_name[variable_name] = {'variable': ov}

            # store the variables in dict by name-type (e.g. "status_code-String")
            self._output_variables_by_type[f'{variable_name}-{variable_type}'] = {'variable': ov}

    def _read(self, key, embedded=True, b64decode=True, decode=False):
        """Create the value in Redis if applicable."""
        if key is None:
            self.log.warning('The key is None.')
            return None

        # get variable type from variable value
        variable_type = self.variable_type(key)

        value = self.key_value_store.read(key.strip())
        if value is None:
            return value

        if variable_type == 'Binary':
            value = self._load_value(value)

            if b64decode:
                value = base64.b64decode(value)
                if decode:
                    value = self._decode_binary(value)
        elif variable_type == 'KeyValue':
            # embedded variable can be unquoted, which breaks JSON.
            value = self._wrap_embedded_keyvalue(value)

            if embedded:
                value = self._read_embedded(value)

            value = self._load_value(value)
        elif variable_type == 'String':
            if embedded:
                value = self._read_embedded(value)

            # coerce string values
            value = self._coerce_string_value(self._load_value(value))
        elif variable_type == 'TCEntity':
            value = self._load_value(value)

        return value

    def _read_array(self, key, embedded=True, b64decode=True, decode=False):
        """Create the value in Redis if applicable."""
        if key is None:  # pragma: no cover
            self.log.warning('The null value for key was provided.')
            return None

        # get variable type from variable value
        variable_type = self.variable_type(key)

        value = self.key_value_store.read(key.strip())
        if value is None:
            return value

        if variable_type == 'BinaryArray':
            value = json.loads(value, object_pairs_hook=OrderedDict)

            values = []
            for v in value:
                if v is not None and b64decode:
                    v = base64.b64decode(v)
                    if decode:
                        v = self._decode_binary(v)
                values.append(v)
            value = values
        elif variable_type == 'KeyValueArray':
            # embedded variable can be unquoted, which breaks JSON.
            value = self._wrap_embedded_keyvalue(value)

            if embedded:
                value = self._read_embedded(value)

            try:
                value = json.loads(value, object_pairs_hook=OrderedDict)
            except ValueError as e:  # pragma: no cover
                raise RuntimeError(f'Failed loading JSON data ({value}). Error: ({e})')
        elif variable_type == 'StringArray':
            if embedded:
                value = self._read_embedded(value)

            # convert int to str
            value_coerced = []
            for v in self._load_value(value):
                # coerce string values
                value_coerced.append(self._coerce_string_value(v))
            value = value_coerced
        elif variable_type in ['TCEntityArray', 'TCEnhancedEntity', 'TCEnhancedEntityArray']:
            value = self._load_value(value)

        # self.log.trace(f'pb create - context: {self._context}, key: {key}, value: {value}')
        return value

    def _read_embedded(self, value):
        """Read method for "embedded" variables.

        .. Note:: The ``read()`` method will automatically determine if the input is a variable or
            needs to be searched for embedded variables.

        Embedded variable rules:

        * Only user input can have embedded variables.
        * Only String and KeyValueArray variables can have embedded variables.
        * Variables can only be embedded one level deep.

        This method will automatically covert variables embedded in a string with value retrieved
        from DB. If there are no keys/variables the raw string will be returned.

        Examples::

            DB Values
            #App:7979:variable_name!String:
                "embedded \\"variable\\""
            #App:7979:two!String:
                "two"
            #App:7979:variable_name!StringArray:
                ["one", "two", "three"]

            Examples 1:
                Input:  "This input has a embedded #App:7979:variable_name!String"

            Examples 2:
                Input: ["one", #App:7979:two!String, "three"]

            Examples 3:
                Input: [{
                    "key": "embedded string",
                    "value": "This input has a embedded #App:7979:variable_name!String"
                }, {
                    "key": "string array",
                    "value": #App:7979:variable_name!StringArray
                }, {
                    "key": "string",
                    "value": #App:7979:variable_name!String
                }]

        Args:
            value (str): The value to parsed and updated from the DB.

        Returns:
            (str): Results retrieved from DB
        """
        if value is None:  # pragma: no cover
            return value

        for variable in (v.group(0) for v in re.finditer(self._variable_parse, str(value))):
            v = self.read(variable)
            self.log.trace(f'embedded variable: {variable}, value: {v}')
            if isinstance(v, (dict, list)):
                v = json.dumps(v)
                # for KeyValueArray with nested dict/list type replace the
                # quoted value to ensure the resulting data is loadable JSON
                value = re.sub(f'"{variable}"', v, value)

            if v is not None:
                # only replace variable if a non-null value is returned from kv store
                # APP-1030 need to revisit this to handle variable references in kv/kvarrays that
                # are None.  Would like to be able to say if value is just the variable reference,
                # sub None value, else insert '' in string.  That would require a kv-specific
                # version of this method that gets the entire list/dict instead of just the string.
                value = re.sub(variable, v, value)
        return value

    @property
    def _variable_pattern(self):
        """Regex pattern to match and parse a playbook variable."""
        variable_pattern = r'#([A-Za-z]+)'  # match literal (#App,#Trigger) at beginning of String
        variable_pattern += r':([\d]+)'  # app id (:7979)
        variable_pattern += r':([A-Za-z0-9_\.\-\[\]]+)'  # variable name (:variable_name)
        variable_pattern += r'!(StringArray|BinaryArray|KeyValueArray'  # variable type (array)
        variable_pattern += r'|TCEntityArray|TCEnhancedEntityArray'  # variable type (array)
        variable_pattern += r'|String|Binary|KeyValue|TCEntity|TCEnhancedEntity'  # variable type
        variable_pattern += r'|(?:(?!String)(?!Binary)(?!KeyValue)'  # non matching for custom
        variable_pattern += r'(?!TCEntity)(?!TCEnhancedEntity)'  # non matching for custom
        variable_pattern += r'[A-Za-z0-9_-]+))'  # variable type (custom)
        return variable_pattern

    @property
    def _variable_array_types(self):
        """Return list of standard playbook array variable types."""
        return [
            'BinaryArray',
            'KeyValueArray',
            'StringArray',
            'TCEntityArray',
            'TCEnhancedEntityArray',
        ]

    @property
    def _variable_single_types(self):
        """Return list of standard playbook single variable types."""
        return [
            'Binary',
            'KeyValue',
            'String',
            'TCEntity',
            'TCEnhancedEntity',
        ]

    @property
    def _variable_types(self):
        """Return list of standard playbook variable typesd."""
        return self._variable_single_types + self._variable_array_types

    def _wrap_embedded_keyvalue(self, data):
        """Wrap keyvalue embedded variable in double quotes.

        Args:
            data (str): The data with embedded variables.

        Returns:
            (str): Results retrieved from DB
        """
        # TODO: need to verify if core still sends improper JSON for KeyValueArrays
        if data is not None:  # pragma: no cover
            variables = []
            for v in re.finditer(self._vars_keyvalue_embedded, data):
                variables.append(v.group(0))

            for var in set(variables):  # recursion over set to handle duplicates
                # pull (#App:1441:embedded_string!String) from (": #App:1441:embedded_string!String)
                variable_string = re.search(self._variable_parse, var).group(0)
                # reformat to replace the correct instance only, handling the case where a variable
                # is embedded multiple times in the same key value array.
                data = data.replace(var, f'": "{variable_string}"')
        return data

    @property
    def key_value_store(self):
        """Return the correct KV store for this execution."""
        if self._key_value_store is None:
            if self.tcex.default_args.tc_playbook_db_type == 'Redis':
                from ..key_value_store import KeyValueRedis

                self._key_value_store = KeyValueRedis(self._context, self.tcex.redis_client)
            elif self.tcex.default_args.tc_playbook_db_type == 'TCKeyValueAPI':
                from ..key_value_store import KeyValueApi

                self._key_value_store = KeyValueApi(self.tcex.session)
            else:  # pragma: no cover
                raise RuntimeError(
                    f'Invalid DB Type: ({self.tcex.default_args.tc_playbook_db_type})'
                )
        return self._key_value_store

    def create_raw(self, key, value):
        """Create method of CRUD operation for raw data.

        ..important:: Raw data can only be a byte, str or int. Other data structures
            (dict, list, etc) must be serialized.

        Args:
            key (str): The variable to write to the DB.
            value (bytes|int|string): The data to write to the DB.

        Returns:
            (str): Result of DB write.
        """
        data = None
        if key is not None and value is not None:
            data = self.key_value_store.create(key.strip(), value)
        else:
            self.log.warning('The key or value field was None.')
        return data

    def read_raw(self, key):
        """Read method of CRUD operation for raw data.

        ..important:: Bytes input will be returned a as string as there is
            no way to determine data from redis originated as bytes or string.

        Args:
            key (str): The variable to read from the DB.

        Returns:
            (str): Results retrieved from DB.
        """
        value = None
        if key is not None:
            value = self.key_value_store.read(key.strip())
        else:
            self.log.warning('The key field was None.')
        return value

    def parse_variable(self, variable):  # pragma: no cover
        """Set placeholder for child method."""
        raise NotImplementedError('Implemented in child class')

    def read(self, key, array=False, embedded=True):  # pragma: no cover
        """Set placeholder for child method."""
        raise NotImplementedError('Implemented in child class')

    def variable_type(self, variable):  # pragma: no cover
        """Set placeholder for child method."""
        raise NotImplementedError('Implemented in child class')



"""ThreatConnect TI Signature"""
from ..group import Group


class Signature(Group):
    """Unique API calls for Signature API Endpoints

    Valid file_types:
    + Snort
    + Suricata
    + YARA
    + ClamAV
    + OpenIOC
    + CybOX
    + Bro
    + Regex
    + SPL

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        name (str, kwargs): [Required for Create] The name for this Group.
        owner (str, kwargs): The name for this Group. Default to default Org when not provided
        file_name (str, kwargs): The name for the attached signature for this Group.
        file_type (str, kwargs): The signature type for this Group.
        file_text (str, kwargs): The signature content for this Group.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties."""
        super().__init__(
            tcex, sub_type='Signature', api_entity='signature', api_branch='signatures', **kwargs
        )

    def download(self):
        """Download the signature.

        Returns:
            obj: The Request response of the download request.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.download(self.api_type, self.api_branch, self.unique_id)



"""ReadArg App Decorators."""
# standard library
from typing import Callable, List, Union

# third-party
import wrapt

# first-party
from tcex.validators import (
    ValidationError,
    equal_to,
    greater_than,
    greater_than_or_equal,
    in_range,
    less_than,
    less_than_or_equal,
    not_in,
    to_bool,
    to_float,
    to_int,
)


class ReadArg:
    """Read value of App arg resolving any playbook variables.

    This decorator will read the args from Redis, if it is a playbook variable, and pass the
    resolved value to the function.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        @ReadArg(arg='color', default='red')
        @ReadArg(arg='number')
        @ReadArg(
            arg='fruits',
            array=True,
            fail_on=[None, []],
            fail_enabled=True
        )
        def my_method(self, color, number, fruits):
            print('color', color)
            print('number', number)
            print('fruit', fruit)

        # ** OR **

        @ReadArg(arg='color', default='red')
        @ReadArg(arg='number')
        @ReadArg(
            arg='fruits',
            array=True,
            fail_on=[None, []],
            fail_enabled=True
        )
        def my_method(self, **kwargs):
            print('color', kwargs.get('color'))
            print('number', kwargs.get('number')
            print('fruit', kwargs.get('fruit')

    Args:
        arg (str): The arg name from the App which contains the input. This input can be
            a Binary, BinaryArray, KeyValue, KeyValueArray, String, StringArray, TCEntity, or
            TCEntityArray.
        array (bool, kwargs): Defaults to False. If True the arg value will always be
            returned as an array.
        default (str, kwargs): Defaults to None. Default value to pass to method if arg
            value is None. Only supported for String or StringArray.
        equal_to (Union[Any, List[Any], Dict[str, Any], kwargs):  Verifies that the arg value is
            equal to the given compare_to value.
        fail_enabled (boolean|str, kwargs): Accepts a boolean or string value.  If a boolean value
            is provided that value will control enabling/disabling this feature. A string
            value should reference an item in the args namespace which resolves to a boolean.
            The value of this boolean will control enabling/disabling this feature.
        fail_on (list, kwargs): Defaults to None. Fail if data read from Redis is in list.
        greater_than (Union[Any, List[Any], Dict[str, Any], kwargs):  Verifies that the arg value is
            greater than the given compare_to value.
        greater_than_or_equal (Union[Any, List[Any], Dict[str, Any], kwargs):  Verifies that the arg
            value is greater than or equal to the given compare_to value.
        group_values (bool, kwargs): Defaults to False. If True, return a list of group names from
            the given argument.
        group_ids (bool, kwargs): Defaults to False. If True, return a list of group ids from the
            given argument.
        in_range (Dict[str, float]|List[float], kwargs): Validator that verifies the arg value is in
            the range min <= value <= max.  For Array args, will verify each member of the array.
            Expected value is {'min': <min_value>, 'max': <max_value>} OR [<min_value>, <max_value>]
        indicator_values (bool, kwargs): Defaults to False. If True, return a list of
            indicator values from the given argument (e.g. ["foo.example.com", "bar.example.com"]).
        less_than (Union[Any, List[Any], Dict[str, Any], kwargs):  Verifies that the arg value is
            less than the given compare_to value.
        less_than_or_equal (Union[Any, List[Any], Dict[str, Any], kwargs):  Verifies that the arg
            value is less than or equal to the given compare_to value.
        strip_values (bool, kwargs): Defaults to False.  If True, automatically strips whitespace
            from arg value.  Not valid if group_values, group_ids, or indicator_values is True.
        to_bool (bool, kwargs): Transform the arg value into a bool where 'true' (case-insensitive),
            '1' (str), and 1 (int or float) will become True and everything else will become False.
            Accepts one parameter 'disallow_none' which, if True, will fail if the arg value is
            none.
        to_float (bool, kwargs):  Transform the arg value into an float.  Accepts one
            parameter 'disallow_none' which, if True, will fail if the arg value is none.
        to_int (bool, kwargs):  Transform the arg value into an int.  Accepts one
            parameter 'disallow_none' which, if True, will fail if the arg value is none.
        validators (Union[List[Callable], Callable]): A callable or list of callables that will be
            invoked with the arg value and arg name and must raise a ValidationError if
            the arg is not valid.  Any return value will be ignored.  Validators will be invoked in
            the order they are given.  If one validator fails, no others will be invoked.
    """

    def __init__(self, arg, **kwargs):
        """Initialize Class properties"""
        self.arg = arg
        self.array = kwargs.get('array')
        self.default = kwargs.get('default')
        self.fail_enabled = kwargs.get('fail_enabled', True)
        self.fail_msg = kwargs.get('fail_msg')
        self.fail_on = kwargs.get('fail_on', [])
        self.embedded = kwargs.get('embedded', True)
        self.indicator_values = kwargs.get('indicator_values', False)
        self.group_values = kwargs.get('group_values', False)
        self.group_ids = kwargs.get('group_ids', False)
        self.strip_values = kwargs.get('strip_values', False)
        self.transforms: Union[List[Callable], Callable] = kwargs.get('transforms', [])
        self.validators: Union[List[Callable], Callable] = kwargs.get('validators', [])
        if self.fail_on:
            self.validators.insert(0, not_in(self.fail_on))

        self._init_validators(**kwargs)

    def _init_validators(self, **kwargs):
        validators_map = {
            'in_range': in_range,
            'equal_to': equal_to,
            'less_than': less_than,
            'less_than_or_equal': less_than_or_equal,
            'greater_than': greater_than,
            'greater_than_or_equal': greater_than_or_equal,
        }

        for k, v in filter(lambda item: item[0] in validators_map, kwargs.items()):
            validator = validators_map.get(k)
            if isinstance(v, list):
                self.validators.append(validator(*v))
            elif isinstance(v, dict):
                self.validators.append(validator(**v))
            else:
                self.validators.append(validator(v))

        transform_map = {
            'to_int': to_int,
            'to_float': to_float,
            'to_bool': to_bool,
        }

        for transform_name in filter(lambda t: t in transform_map, kwargs.keys()):
            transform = transform_map.get(transform_name)
            transform_args = kwargs.get(transform_name)
            if isinstance(transform_args, list):
                self.transforms.append(transform(*transform_args))
            elif isinstance(transform_args, dict):
                self.transforms.append(transform(**transform_args))
            else:
                self.transforms.append(transform())

    @wrapt.decorator
    def __call__(self, wrapped, instance, args, kwargs):
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """

        def read(app, *args, **kwargs):
            """Retrieve/Read data from Redis.

            Args:
                app (class): The instance of the App class "self".
            """
            # retrieve the label for the current Arg
            label = app.tcex.ij.params_dict.get(self.arg, {}).get('label')

            # self.enable (e.g., True or 'fail_on_false') enables/disables this feature
            enabled = self.fail_enabled
            if not isinstance(self.fail_enabled, bool):
                enabled = getattr(app.args, self.fail_enabled)
                if not isinstance(enabled, bool):  # pragma: no cover
                    raise RuntimeError(
                        'The fail_enabled value must be a boolean or resolved to bool.'
                    )
                app.tcex.log.debug(f'Fail enabled is {enabled} ({self.fail_enabled}).')

            try:
                # read arg from namespace
                arg = getattr(app.args, self.arg)
            except AttributeError:
                if enabled:
                    app.tcex.log.error(f'Arg {self.arg} was not found in Arg namespace.')
                    message = self.fail_msg or f'Invalid value provided for "{label}" ({self.arg}).'
                    app.exit_message = message  # for test cases
                    app.tcex.exit(1, message)
                else:
                    # add results to kwargs
                    kwargs[self.arg] = self.default
                    return wrapped(*args, **kwargs)

            # retrieve data from Redis and call decorated function
            if self.indicator_values:
                arg_data = app.tcex.playbook.read_indicator_values(arg)
            elif self.group_values:
                arg_data = app.tcex.playbook.read_group_values(arg)
            elif self.group_ids:
                arg_data = app.tcex.playbook.read_group_ids(arg)
            else:
                arg_data = app.tcex.playbook.read(arg, self.array, embedded=self.embedded)
                if self.strip_values and arg_data is not None:
                    variable_type = app.tcex.playbook.variable_type(arg)
                    if variable_type == 'String':
                        arg_data = arg_data.strip()
                    elif variable_type == 'StringArray':
                        arg_data = [s.strip() for s in arg_data]
                    else:
                        app.tcex.log.warn(
                            f'strip_values is enabled but variable is of type {variable_type}, '
                            'only String and StringArray variables can be stripped.'
                        )

            if self.default is not None and arg_data is None:
                arg_data = self.default
                app.tcex.log.debug(
                    f'replacing null value with provided default value "{self.default}".'
                )

            try:
                for transform in self.transforms:
                    arg_data = transform(arg_data, self.arg, label)
            except ValidationError as v:
                value_formatted = f'"{arg_data}"' if isinstance(arg_data, str) else str(arg_data)
                message = f'Invalid value ({value_formatted}) found for "{label}": {v.message}'
                app.tcex.log.error(message)
                if self.fail_msg:
                    app.exit_message = self.fail_msg  # for test cases
                    app.tcex.exit(1, self.fail_msg)
                else:
                    app.exit_message = message
                    app.tcex.exit(1, message)

            # check arg_data against fail_on_values
            if enabled:
                try:
                    list([v(arg_data, self.arg, label) for v in self.validators])
                except ValidationError as v:
                    value_formatted = (
                        f'"{arg_data}"' if isinstance(arg_data, str) else str(arg_data)
                    )
                    message = f'Invalid value ({value_formatted}) found for "{label}": {v.message}'
                    app.tcex.log.error(message)
                    if self.fail_msg:
                        app.exit_message = self.fail_msg  # for test cases
                        app.tcex.exit(1, self.fail_msg)
                    else:
                        app.exit_message = message
                        app.tcex.exit(1, message)
            # add results to kwargs
            kwargs[self.arg] = arg_data

            # Add logging for debug/troubleshooting
            arg_type = app.tcex.playbook.variable_type(getattr(app.args, self.arg))
            if arg_type not in ['Binary', 'BinaryArray'] and app.tcex.log.getEffectiveLevel() <= 10:
                # only log variable data to prevent logging keychain data
                if app.tcex.playbook.is_variable(arg):
                    log_string = str(arg_data)
                    if len(log_string) > 100:  # pragma: no cover
                        log_string = f'{log_string[:100]} ...'
                    app.tcex.log.debug(f'input value: {log_string}')

            return wrapped(*args, **kwargs)

        return read(instance, *args, **kwargs)



"""TcEx Framework Key Value Redis Module"""


class KeyValueRedis:
    """TcEx Key Value Redis Module.

    Args:
        context (str): The Redis context (hash) for hashed based operations.
        redis_client (redis.Client): An instance of redis client.
    """

    def __init__(self, context, redis_client):
        """Initialize the Class properties."""
        self._context = context
        self._redis_client = redis_client

    @property
    def context(self):
        """Return the current context."""
        return self._context

    @context.setter
    def context(self, context):
        """Set or update the current context."""
        self._context = context

    def create(self, key, value):
        """Create key/value pair in Redis.

        Args:
            key (str): The field name (key) for the kv pair in Redis.
            value (any): The value for the kv pair in Redis.

        Returns:
            str: The response from Redis.
        """
        return self._redis_client.hset(self.context, key, value)

    def delete(self, key):
        """Alias for hdel method.

        Args:
            key (str): The field name (key) for the kv pair in Redis.

        Returns:
            str: The response from Redis.
        """
        return self._redis_client.hdel(self.context, key)

    def hgetall(self):
        """Read data from Redis for the current context.

        Returns:
            list: The response data from Redis.
        """
        return self._redis_client.hgetall(self.context)

    def read(self, key):
        """Read data from Redis for the provided key.

        Returns:
            str: The response data from Redis.
        """
        value = self._redis_client.hget(self.context, key)
        # convert retrieved bytes to string
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        return value



"""ThreatConnect Note"""
from .api_endpoints import ApiEndpoints
from .common_case_management import CommonCaseManagement
from .common_case_management_collection import CommonCaseManagementCollection
from .filter import Filter
from .tql import TQL


class Notes(CommonCaseManagementCollection):
    """ThreatConnect Notes Object

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # Example of params input
        {
            'result_limit': 100,  # How many results are retrieved.
            'result_start': 10,  # Starting point on retrieved results.
            'fields': ['caseId', 'summary']  # Additional fields returned on the results
        }

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        initial_response (dict, optional): Initial data in
            Case Object for Note. Defaults to None.
        tql_filters (list, optional): List of TQL filters. Defaults to None.
        params(dict, optional): Dict of the params to be sent while
            retrieving the Notes objects.
    """

    def __init__(self, tcex, initial_response=None, tql_filters=None, params=None):
        """Initialize Class properties."""
        super().__init__(
            tcex,
            ApiEndpoints.NOTES,
            initial_response=initial_response,
            tql_filters=tql_filters,
            params=params,
        )
        if initial_response:
            for item in initial_response.get('data', []):
                self.added_items.append(Note(tcex, **item))

    def __iter__(self):
        """Object iterator"""
        return self.iterate(initial_response=self.initial_response)

    def add_note(self, note):
        """Add a Note.

        Args:
            note (case_management.note.Note): The Note Object to add.
        """
        self.added_items.append(note)

    def entity_map(self, entity):
        """Update current object with provided object properties.

        Args:
            entity (dict): The dict to map self too.
        """
        return Note(self.tcex, **entity)

    @property
    def filter(self):
        """Return instance of FilterNotes Object."""
        return FilterNotes(ApiEndpoints.NOTES, self.tcex, self.tql)


class Note(CommonCaseManagement):
    """Note object for Case Management.

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        artifact (Artifact, kwargs): [Read-Only] The **Artifact** for the Note.
        artifact_id (int, kwargs): the ID of the Artifact on which to apply the Note
        author (int, kwargs): [Read-Only] The **Author** for the Note.
        case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Note.
        case_xid (str, kwargs): [Required (alt: caseId)]
        date_added (str, kwargs): [Read-Only] The **Date Added** for the Note.
        edited (boolean, kwargs): [Read-Only] The **Edited** for the Note.
        last_modified (str, kwargs): [Read-Only] The **Last Modified** for the Note.
        parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Note.
        summary (str, kwargs): [Read-Only] The **Summary** for the Note.
        task (case_management.task.Task, kwargs): [Read-Only] The **Task** for the Note.
        task_id (int, kwargs): the ID of the Task on which to apply the Note
        task_xid (str, kwargs): the XID of the Task on which to apply the Note
        text (str, kwargs): [Required] The **Text** for the Note.
        workflow_event (WorkflowEvent, kwargs): [Read-Only] The **Workflow Event** for the Note.
        workflow_event_id (int, kwargs): the ID of the Event on which to apply the Note
    """

    def __init__(self, tcex, **kwargs):
        """Initialize class properties."""
        super().__init__(tcex, ApiEndpoints.NOTES, kwargs)

        self._artifact = kwargs.get('artifact', None)
        self._artifact_id = kwargs.get('artifact_id', None)
        self._author = kwargs.get('author', None)
        self._case_id = kwargs.get('case_id', None)
        self._case_xid = kwargs.get('case_xid', None)
        self._date_added = kwargs.get('date_added')
        self._edited = kwargs.get('edited', None)
        self._last_modified = kwargs.get('last_modified', None)
        self._parent_case = kwargs.get('parent_case', None)
        self._summary = kwargs.get('summary', None)
        self._task = kwargs.get('task', None)
        self._task_id = kwargs.get('task_id', None)
        self._task_xid = kwargs.get('task_xid', None)
        self._text = kwargs.get('text', None)
        self._workflow_event = kwargs.get('workflow_event', None)
        self._workflow_event_id = kwargs.get('workflow_event_id', None)

    @property
    def artifact(self):
        """Return the parent **Artifact** for the Note."""
        if self._artifact:
            return self.tcex.cm.artifact(**self._artifact)
        return self._artifact

    @property
    def artifact_id(self):
        """Return the parent **Artifact ID** for the Note."""
        return self._artifact_id

    @artifact_id.setter
    def artifact_id(self, artifact_id):
        """Set the parent **Artifact ID** for the Note."""
        self._artifact_id = artifact_id

    @property
    def as_entity(self):
        """Return the entity representation of the Note."""
        return {'type': 'Note', 'value': self.summary, 'id': self.id}

    @property
    def author(self):
        """Return the parent **Author** for the Note."""
        return self._author

    def entity_mapper(self, entity):
        """Update current object with provided object properties.

        Args:
            entity (dict): An entity dict used to update the Object.
        """
        new_note = Note(self.tcex, **entity)
        self.__dict__.update(new_note.__dict__)

    @property
    def case_id(self):
        """Return the parent **Case ID** for the Note."""
        return self._case_id

    @case_id.setter
    def case_id(self, case_id):
        """Set the parent **Case ID** for the Note."""
        self._case_id = case_id

    @property
    def case_xid(self):
        """Return the parent **Case XID** for the Note."""
        return self._case_xid

    @case_xid.setter
    def case_xid(self, case_xid):
        """Set the parent **Case XID** for the Note."""
        self._case_xid = case_xid

    @property
    def date_added(self):
        """Return the **Date Added** value for the Note."""
        return self._date_added

    @property
    def edited(self):
        """Return the **Edited** value for the Note."""
        return self._edited

    @property
    def last_modified(self):
        """Return the **Last Modified** value for the Note."""
        return self._last_modified

    @property
    def parent_case(self):
        """Return the **Parent Case** for the Task."""
        if self._parent_case:
            return self.tcex.cm.case(**self._parent_case)
        return self._parent_case

    @property
    def summary(self):
        """Return the **Summary** value for the Note."""
        return self._summary

    @property
    def task(self):
        """Return the **Task** for the Note."""
        if self._task:
            return self.tcex.cm.task(**self._task)
        return self._task

    @property
    def task_id(self):
        """Return the **Task ID** for the Note."""
        return self._task_id

    @task_id.setter
    def task_id(self, task_id):
        """Set the **Task ID** for the Note."""
        self._task_id = task_id

    @property
    def task_xid(self):
        """Return the **Task XID** for the Note."""
        return self._task_xid

    @task_xid.setter
    def task_xid(self, task_xid):
        """Set the **Task XID** for the Note."""
        self._task_xid = task_xid

    @property
    def text(self):
        """Return the **Text** value for the Note."""
        return self._text

    @text.setter
    def text(self, text):
        """Set the **Text** value for the Note."""
        self._text = text

    @property
    def workflow_event(self):
        """Return the **Workflow Event** for the Note."""
        if self._workflow_event:
            return self.tcex.cm.workflow_event(**self._workflow_event)
        return self._workflow_event

    @property
    def workflow_event_id(self):
        """Return the **Workflow Event ID** value for the Note."""
        return self._workflow_event_id

    @workflow_event_id.setter
    def workflow_event_id(self, workflow_event_id):
        """Set the **Workflow Event ID** value for the Note."""
        self._workflow_event_id = workflow_event_id


class FilterNotes(Filter):
    """Filter Object for Notes"""

    def artifact_id(self, operator, artifact_id):
        """Filter Notes based on **artifactId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            artifact_id (int): The ID of the artifact this note is associated with.
        """
        self._tql.add_filter('artifactId', operator, artifact_id, TQL.Type.INTEGER)

    def author(self, operator, author):
        """Filter Notes based on **author** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            author (str): The account login of the user who wrote the note.
        """
        self._tql.add_filter('author', operator, author, TQL.Type.STRING)

    def case_id(self, operator, case_id):
        """Filter Notes based on **caseId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            case_id (int): The ID of the case this note is associated with.
        """
        self._tql.add_filter('caseId', operator, case_id, TQL.Type.INTEGER)

    def date_added(self, operator, date_added):
        """Filter Notes based on **dateAdded** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            date_added (str): The date the note was written.
        """
        self._tql.add_filter('dateAdded', operator, date_added, TQL.Type.STRING)

    @property
    def has_artifact(self):
        """Return **FilterArtifacts** for further filtering."""
        from .artifact import FilterArtifacts

        artifacts = FilterArtifacts(ApiEndpoints.ARTIFACTS, self._tcex, TQL())
        self._tql.add_filter('hasArtifact', TQL.Operator.EQ, artifacts, TQL.Type.SUB_QUERY)
        return artifacts

    @property
    def has_case(self):
        """Return **FilterCases** for further filtering."""
        from .case import FilterCases

        cases = FilterCases(ApiEndpoints.CASES, self._tcex, TQL())
        self._tql.add_filter('hasCase', TQL.Operator.EQ, cases, TQL.Type.SUB_QUERY)
        return cases

    @property
    def has_task(self):
        """Return **FilterTask** for further filtering."""
        from .task import FilterTasks  # pylint: disable=cyclic-import

        tasks = FilterTasks(ApiEndpoints.TASKS, self._tcex, TQL())
        self._tql.add_filter('hasTask', TQL.Operator.EQ, tasks, TQL.Type.SUB_QUERY)
        return tasks

    def id(self, operator, id):  # pylint: disable=redefined-builtin
        """Filter Notes based on **id** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            id (int): The ID of the case.
        """
        self._tql.add_filter('id', operator, id, TQL.Type.INTEGER)

    def last_modified(self, operator, last_modified):
        """Filter Notes based on **lastModified** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            last_modified (str): The date the note was last modified.
        """
        self._tql.add_filter('lastModified', operator, last_modified, TQL.Type.STRING)

    def summary(self, operator, summary):
        """Filter Notes based on **summary** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            summary (str): Text of the first 100 characters of the note.
        """
        self._tql.add_filter('summary', operator, summary, TQL.Type.STRING)

    def task_id(self, operator, task_id):
        """Filter Notes based on **taskId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            task_id (int): The ID of the task this note is associated with.
        """
        self._tql.add_filter('taskId', operator, task_id, TQL.Type.INTEGER)

    def workflow_event_id(self, operator, workflow_event_id):
        """Filter Notes based on **workflowEventId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            workflow_event_id (int): The ID of the workflow event this note is associated with.
        """
        self._tql.add_filter('workflowEventId', operator, workflow_event_id, TQL.Type.INTEGER)



"""ThreatConnect Common Case Management"""
# third-party
from requests.exceptions import ProxyError

# first-party
from tcex.case_management.common_case_management_collection import CommonCaseManagementCollection


class CommonCaseManagement:
    """Common Class for Case Management.

    Args:
        tcex (TcEx): An instance of tcex.
        api_endpoint (str): The path to the API endpoint.
    """

    def __init__(self, tcex, api_endpoint, kwargs):
        """Initialize Class properties."""
        self.tcex = tcex
        self.api_endpoint = api_endpoint.value

        # properties
        self._fields = None
        self._id = kwargs.get('id', None)
        self._properties = None
        self._tql = None

        # process kwargs
        self._transform_kwargs(kwargs)

    def __str__(self):
        """Printable version of Object"""
        printable_string = ''
        for key, value in sorted(vars(self).items()):
            key = key.lstrip('_')
            if key in self._excluded_properties:
                continue

            # format key with trailing whitespace
            key = f'{key} '

            if isinstance(value, (int, str)) or value is None:
                printable_string += f'{key:.<40} {str(value):<50}\n'
            else:
                printable_string += f'{key:.<40} {"<object>":<50}\n'
        return printable_string

    def _camel_case_key(self, key):
        """Map key to it's appropriate Core/Test value."""
        if key in ['staged_artifact_type', 'stagedArtifactType']:  # pragma: no cover
            # used in testing when staging data
            return 'type'
        return self.tcex.utils.camel_to_snake(key)

    @property
    def _doc_string(self):
        """Return a doc string current object."""
        docstring = (
            f'\n{" " * 4}"""{self.__class__.__name__} object for Case Management.\n\n'
            f'{" " * 4}Args:\n'
            f'{" " * 8}tcex (TcEx): An instantiated instance of TcEx object.\n'
        )
        for p, pd in sorted(self.properties.items()):
            # get data array if exists
            data = pd.get('data')

            # get read-only value for display and required value
            read_only = pd.get('read-only', False)
            if read_only is None and data:
                read_only = data[0].get('read-only', False)

            # get required value or default
            required = pd.get('required', False)
            # if read_only is True:
            #     required = False

            # get type
            type_flag = ''  # used to append the word flag for bool types
            type_ = pd.get('type')
            if type_ is None and data:
                type_ = data[0].get('type')

            if type_ == 'Array':
                type_ = 'list'
            elif type_ == 'Boolean':
                type_ = 'bool'
                type_flag = ' flag'
            elif type_ == 'Date':
                type_ = 'str'
            elif type_ == 'Integer':
                type_ = 'int'
            elif type_ == 'JsonNode':
                type_ = 'dict'
            elif type_ == 'Long':
                type_ = 'int'
            elif type_ == 'String':
                type_ = 'str'

            # set description
            description = (
                f'The **{self.tcex.utils.camel_to_space(p).title()}**{type_flag} for the '
                f'{self.tcex.utils.camel_to_space(self.__class__.__name__).title()}.'
            )
            if pd.get('description'):
                description = pd.get('description')
            elif data:
                if data[0].get('description'):
                    description = data[0].get('description', '')

            # build meta with required and read-only values
            meta = []
            if required:
                required_text = 'Required'
                if pd.get('required-alt-field'):
                    required_text += f" (alt: {pd.get('required-alt-field')})"
                meta.append(required_text)
            if read_only:
                meta.append('Read-Only')
            if meta:
                meta = f" [{','.join(meta)}]"
            else:
                meta = ''

            # build docstring
            docstring += (
                f'{" " * 8}{self.tcex.utils.camel_to_snake(p)} ({type_}, kwargs):'
                f'{meta} {description}\n'
            )

        docstring += f'{" " * 4}"""\n'
        return docstring

    @property
    def _excluded_properties(self):
        """Return a list of properties to exclude when creating a dict of the children class."""
        return [
            'api_endpoint',
            'artifact_filter',
            'case_filter',
            'filter',
            'fields',
            'kwargs',
            'put_properties',
            'post_properties',
            'properties',
            'task_filter',
            'tcex',
            'tql',
            'workflow_event_filter',
        ]

    def _reverse_transform(self, kwargs):
        """Reverse mapping of the _metadata_map method."""

        def reverse_transform(current_kwargs):
            reversed_mappings = {}
            for key, value in current_kwargs.items():
                if isinstance(value, dict):
                    value = reverse_transform(value)
                elif isinstance(value, list):
                    values = []
                    for entry in value:
                        values.append(self._reverse_transform(entry))
                    reversed_mappings[key] = values
                    return reversed_mappings
                new_key = self.tcex.utils.snake_to_camel(key)
                reversed_mappings[new_key] = value
            return reversed_mappings

        return reverse_transform(kwargs)

    def _transform_kwargs(self, kwargs):
        """Map the provided kwargs to expected arguments."""
        for key in dict(kwargs):
            kwargs[self._camel_case_key(key)] = kwargs.pop(key)

    @property
    def as_entity(self):  # pragma: no cover
        """Return the object as an entity."""
        raise NotImplementedError('Child class must implement this method.')

    @property
    def as_dict(self):
        """Return the dict representation of the CM object."""
        properties = vars(self)
        as_dict = {}
        for key, value in properties.items():
            key = key.lstrip('_')
            if key in self._excluded_properties:
                continue
            if hasattr(self, key):
                value = getattr(self, key)
            try:
                value = value.as_dict
            except AttributeError:
                if isinstance(value, CommonCaseManagementCollection):
                    for added_item in value.added_items:
                        if key not in as_dict:
                            as_dict[key] = {'data': []}
                        as_dict[key]['data'].append(added_item.as_dict)
                    continue

            if value is None:
                continue
            if isinstance(value, dict) and 'data' in value and not value.get('data'):
                continue
            as_dict[key] = value

        # don't return empty dicts
        if not as_dict:
            return None

        return as_dict

    @property
    def body(self):
        """Return the body representation of the CM object."""
        properties = vars(self)
        built_body = {}
        for key, value in properties.items():
            key = key.lstrip('_')
            if key in self._excluded_properties:
                continue
            # If creating a PUT body and key is not a PUT property do not include it in body.
            if self.id and key not in map(self._camel_case_key, self.put_properties):
                continue
            # If creating a POST body and key is not a POST property do not include it in body.
            if key not in map(self._camel_case_key, self.post_properties):
                continue
            if hasattr(self, key):
                value = getattr(self, key)
            try:
                value = value.body
            except AttributeError:
                if isinstance(value, CommonCaseManagementCollection):
                    for added_item in value.added_items:
                        if key not in built_body:
                            built_body[key] = {'data': []}
                        built_body[key]['data'].append(added_item.body)
                    continue

            if value is None:
                continue
            if isinstance(value, dict) and 'data' in value and not value.get('data'):
                continue
            built_body[key] = value

        # don't return empty dicts
        if not built_body:
            return None

        return self._reverse_transform(built_body)

    @property
    def available_fields(self):
        """Return the available query param field names for this object."""
        return [fd.get('name') for fd in self.fields]

    @property
    def fields(self):
        """Return the field data for this object."""
        if self._fields is None:
            r = self.tcex.session.options(f'{self.api_endpoint}/fields', params={})
            if r.ok:
                self._fields = r.json()['data']
        return self._fields

    def delete(self):
        """Delete the Case Management Object.

        If no id is present in the obj then returns immediately.
        """
        if not self.id:  # pragma: no cover
            self.tcex.log.warning('A case without an ID cannot be deleted.')
            return

        url = f'{self.api_endpoint}/{self.id}'
        r = None
        try:
            r = self.tcex.session.delete(url)
            self.tcex.log.debug(
                f'Method: ({r.request.method.upper()}), '
                f'Status Code: {r.status_code}, '
                f'URl: ({r.url})'
            )
        except (ConnectionError, ProxyError):  # pragma: no cover
            self.tcex.handle_error(
                951, ['OPTIONS', 407, '{\"message\": \"Connection Error\"}', self.api_endpoint]
            )
        if len(r.content) < 5000:
            self.tcex.log.debug(f'response text: {r.text}')
        else:  # pragma: no cover
            self.tcex.log.debug('response text: (text to large to log)')
        if not self.success(r):
            err = r.text or r.reason
            if r.status_code == 404:
                self.tcex.handle_error(952, [r.request.method.upper(), r.status_code, err, r.url])
            self.tcex.handle_error(950, [r.status_code, err, r.url])
        return

    def entity_mapper(self, entity):  # pragma: no cover
        """Stub for entity mapper"""
        raise NotImplementedError('Child class must implement this method.')

    def get(self, all_available_fields=False, case_management_id=None, params=None):
        """Get the Case Management Object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            all_available_fields (bool): If True all available fields will be returned.
            case_management_id (int): The id of the case management object to be returned.
            retry_count (int, optional): [description]. Defaults to 0.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Artifacts objects.

        Returns:
            Object: A artifact, case, note, task, tag, or workflow object.
        """
        if params is None:
            params = {}

        # convert all keys to camel case
        for k, v in list(params.items()):
            del params[k]
            k = self.tcex.utils.snake_to_camel(k)
            params[k] = v

        cm_id = case_management_id or self.id
        url = f'{self.api_endpoint}/{cm_id}'

        # add fields parameter if provided
        if all_available_fields:
            params['fields'] = []
            for field in self.available_fields:
                params['fields'].append(field)

        if not cm_id:  # pragma: no cover
            message = '{"message": "No ID provided.", "status": "Error"}'
            self.tcex.handle_error(952, ['GET', '404', message, url])

        r = None
        try:
            r = self.tcex.session.get(url, params=params)
        except (ConnectionError, ProxyError):  # pragma: no cover
            self.tcex.handle_error(
                951, ['OPTIONS', 407, '{\"message\": \"Connection Error\"}', self.api_endpoint]
            )
        self.tcex.log.debug(
            f'Method: ({r.request.method.upper()}), '
            f'Status Code: {r.status_code}, '
            f'URl: ({r.url})'
        )
        if len(r.content) < 5000:
            self.tcex.log.debug(f'response text: {r.text}')
        else:  # pragma: no cover
            self.tcex.log.debug('response text: (text to large to log)')
        if not self.success(r):
            err = r.text or r.reason
            if r.status_code == 404:
                self.tcex.handle_error(952, ['GET', r.status_code, err, r.url])
            self.tcex.handle_error(951, ['GET', r.status_code, err, r.url])
        self.entity_mapper(r.json().get('data', {}))
        return r

    @property
    def id(self):
        """Return the id of the case management object."""
        return self._id

    @id.setter
    def id(self, cm_id):
        """Set the id of the case management object."""
        self._id = cm_id

    @property
    def implemented_properties(self):
        """Return implemented Object properties."""
        properties = []
        for key in vars(self):
            key = key.lstrip('_')
            if key in self._excluded_properties:
                continue

            properties.append(key)

        return properties

    @property
    def put_properties(self):
        """Return all the properties available in PUT requests."""
        put_properties = []
        for p, pd in sorted(self.properties.items()):
            # get read-only value for display and required value
            updatable = pd.get('updatable', True)
            read_only = pd.get('read-only', False)
            if not read_only:
                if updatable:
                    put_properties.append(p)

        return put_properties

    @property
    def post_properties(self):
        """Return all the properties available in POST requests."""
        post_properties = []
        for p, pd in sorted(self.properties.items()):
            read_only = pd.get('read-only', False)
            if not read_only:
                post_properties.append(p)

        return post_properties

    @property
    def properties(self):
        """Return defined API properties for the current object."""
        if self._properties is None:
            try:
                r = self.tcex.session.options(self.api_endpoint, params={'show': 'readOnly'})
                if r.ok:
                    self._properties = r.json()
            except (ConnectionError, ProxyError):
                self.tcex.handle_error(
                    951, ['OPTIONS', 407, '{\"message\": \"Connection Error\"}', self.api_endpoint]
                )
        return self._properties

    @property
    def required_properties(self):
        """Return a list of required fields for current object."""
        rp = []
        for p, pd in self.properties.items():
            if pd.get('required'):
                rp.append(p)
        return rp

    def submit(self):
        """Create or Update the Case Management object.

        This is determined based on if the id is already present in the object.
        """
        body = self.body or {}

        method = 'POST'
        url = self.api_endpoint
        if self.id:
            # if the ID is included, its an update
            method = 'PUT'
            url = f'{self.api_endpoint}/{self.id}'

        # make the request
        r = self.tcex.session.request(method, url, json=body)

        self.tcex.log.debug(
            f'Method: ({r.request.method.upper()}), '
            f'Status Code: {r.status_code}, '
            f'URl: ({r.url})'
        )
        self.tcex.log.debug(f'body: {body}')
        if len(r.content) < 5000:
            self.tcex.log.debug(f'response text: {r.text}')
        else:  # pragma: no cover
            self.tcex.log.debug('response text: (text to large to log)')

        if not self.success(r):  # pragma: no cover
            err = r.text or r.reason
            if r.status_code == 404:
                self.tcex.handle_error(952, [r.request.method.upper(), r.status_code, err, r.url])
            self.tcex.handle_error(951, [r.request.method, r.status_code, err, r.url])

        r_json = r.json()
        if not self.id:
            self.id = r_json.get('data', {}).get('id')
        body['id'] = self.id
        self.entity_mapper(r_json.get('data', r_json))

        return r

    @staticmethod
    def success(r):
        """Validate the response is valid.

        Args:
            r (requests.response): The response object.

        Returns:
            bool: True if status is "ok"
        """
        status = True
        if r.ok:
            try:
                if r.json().get('status') != 'Success':  # pragma: no cover
                    status = False
            except Exception:  # pragma: no cover
                status = False
        else:
            status = False
        return status



"""ThreatConnect TI Address"""
from ..indicator import Indicator


class Address(Indicator):
    """Unique API calls for Address API Endpoints"""

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties.

        Args:
            ip (str): The value for this Indicator.
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): [Read-Only] The date timestamp the Indicator was created.
            last_modified (str, kwargs): [Read-Only] The date timestamp the Indicator was last
                modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
        """
        super().__init__(
            tcex, sub_type='Address', api_entity='address', api_branch='addresses', **kwargs
        )
        self.unique_id = kwargs.get('unique_id', kwargs.get('ip'))
        self.data['ip'] = self.unique_id

    def _set_unique_id(self, json_response):
        """Set the unique_id provided a json response.

        Args:
            json_response:
        """
        self.unique_id = json_response.get('ip', '')

    def can_create(self):
        """Return True if address can be created.

        If the ip address has been provided returns that the address can be created, otherwise
        returns that the address cannot be created.
        """
        return not self.data.get('ip') is None

    # TODO: @burdy - is this correct for address?
    def dns_resolution(self):
        """Update the DNS resolution.

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])
        return self.tc_requests.dns_resolution(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner
        )



"""App Decorators Module."""
# standard library
import traceback

# third-party
import wrapt


class OnException:
    """Set exit message on failed execution.

    This decorator will catch the generic "Exception" error, log the supplied error message, set
    the "exit_message", and exit the App with an exit code of 1.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        @OnException(exit_msg='Failed to process JSON data.')
        def my_method(json_data):
            json.dumps(json_data)

    Args:
        exit_msg (str): The message to send to exit method.
        exit_enabled (boolean|str, kwargs): Accepts a boolean or string value.  If a boolean value
            is provided that value will control enabling/disabling this feature. A string
            value should reference an item in the args namespace which resolves to a boolean.
            The value of this boolean will control enabling/disabling this feature.
        write_output (boolean): default True.
            If enabled, will call app.write_output() when an exception is raised.
    """

    def __init__(self, exit_msg=None, exit_enabled=True, write_output=True):
        """Initialize Class properties"""
        self.exit_enabled = exit_enabled
        self.exit_msg = exit_msg or 'An exception has been caught. See the logs for more details.'
        self.write_output = write_output

    @wrapt.decorator
    def __call__(self, wrapped, instance, args, kwargs):
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """

        def exception(app, *args, **kwargs):
            """Call the function and handle any exception.

            Args:
                app (class): The instance of the App class "self".
            """
            # self.enable (e.g., True or 'fail_on_false') enables/disables this feature
            enabled = self.exit_enabled
            if not isinstance(self.exit_enabled, bool):
                enabled = getattr(app.args, self.exit_enabled)
                if not isinstance(enabled, bool):  # pragma: no cover
                    raise RuntimeError(
                        'The exit_enabled value must be a boolean or resolved to bool.'
                    )
                app.tcex.log.debug(f'Fail enabled is {enabled} ({self.exit_enabled}).')

            try:
                return wrapped(*args, **kwargs)
            except Exception:
                app.tcex.log.error(traceback.format_exc())
                app.exit_message = self.exit_msg  # for test cases
                if enabled:
                    if self.write_output:
                        app.tcex.playbook.write_output()
                        if hasattr(app, 'write_output'):
                            app.write_output()
                    app.tcex.exit(1, self.exit_msg)

        return exception(instance, *args, **kwargs)



"""TcEx testing profile Class."""
# standard library
import base64
import pickle  # nosec
import zlib

# third-party
import colorama as c
from requests.sessions import Session

# autoreset colorama
c.init(autoreset=True, strip=False)


class SessionManager:
    """Class for profile Session recording."""

    def __init__(self, profile):
        """Initialize Class properties."""
        self.profile = profile

        # properties
        self.profile_data = self.profile.contents
        self.record = False

    def init(self):
        """Initialize session recording/playback.

        Configured ON with the --record_session test flag, forcibly disabled with the
        --ignore_session test flag. The profile field options.session.enabled can be
        true to enable session recording/playback.

        The profile field options.session.blur may be a list of fields to blur to force
        matching (ie, date/times, passwords, etc)
        """
        if self.profile.pytest_args.get('ignore_session'):
            # return if cli arg passed
            return

        # ensure a profile has all required sections
        self.profile_data.setdefault('stage', {})
        self.profile_data.setdefault('options', {})
        self.profile_data['options'].setdefault('session', {})
        self.profile_data['options']['session'].setdefault('enabled', False)
        self.profile_data['options']['session'].setdefault('blur', [])

        session_options = self.profile_data['options']['session']
        session_enabled = session_options['enabled']

        if 'session' in self.profile_data.get('stage', {}) and not session_enabled:
            # session in staging mean enabled should be true?
            session_enabled = True

        if self.profile.pytest_args.get('record_session'):
            session_enabled = True
            self.profile_data['stage'].setdefault('session', {})
            self.record = True

        self.profile.log.data(
            'session', 'Session Manager', f'session recording enabled: {session_enabled}', 'info',
        )
        if not session_enabled:
            # no setting that would enable recording found so stop here
            return

        # if stage.session doesn't exist, but session_enabled is true, implicitly turn
        # on session recording (someone zapped the data out of the profile)
        if 'session' not in self.profile_data['stage']:
            self.profile_data['stage']['session'] = {}
            self.record = True

        # default blur value
        blur = ['password']
        blur.extend(session_options['blur'])
        self.profile.log.data(
            'session', 'Session Manager', f'bluring params {blur}', 'info',
        )

        # get requests Session attr for monkeypatch
        _request = getattr(Session, 'request')

        # local variable scope for class method/properties to be used in monkeypatch
        session_data = self.profile_data['stage']['session']
        session_manager = self

        # log record value
        self.profile.log.data(
            'session', 'Session Manager', f'record enabled: {self.record}', 'info',
        )

        # Monkeypatch method for requests.sessions.Session.request
        def request(self, method, url, *args, **kwargs):
            """Intercept method for Session.request."""
            params = kwargs.get('params') or {}
            parm_list = []
            params_keys = sorted(params.keys())
            for key in params_keys:
                if key in blur:
                    value = '***'
                else:
                    value = params.get(key)
                parm_list.append((key, value))

            # The key for this request e.g. GET https://... ('foo':'bla')
            request_key = f'{method} {url} {parm_list}'

            # if not recording, we must be playing back
            if not session_manager.record:
                result_data = session_data.get(request_key, None)
                if result_data is None:
                    raise KeyError(f'No stage.session value found for key {request_key}')
                return session_manager.unpickle_result(result_data)

            result = _request(self, method, url, *args, **kwargs)
            pickled_result = session_manager.pickle_result(result)
            session_data[request_key] = pickled_result
            return result

        # Add the intercept
        self.profile.monkeypatch.setattr(Session, 'request', request)

    @staticmethod
    def pickle_result(result):
        """Pickled the result object so we can reconstruct it later"""
        return base64.b64encode(zlib.compress(pickle.dumps(result))).decode('utf-8')

    @staticmethod
    def unpickle_result(result):
        """Reverse the pickle operation"""
        return pickle.loads(zlib.decompress(base64.b64decode(result.encode('utf-8'))))  # nosec

    def update_profile(self):
        """Write back the profile *if* we recorded session data"""
        # only record if _record set to true in init method
        if not self.record:
            return

        # get stage and session data
        stage = self.profile_data.get('stage', {})
        session = stage.get('session', {})

        # update data
        self.profile_data['stage']['session'] = session
        options = self.profile_data.get('options', {})
        self.profile_data['options'] = options
        options['session'] = self.profile_data.get('options').get('session')

        self.profile.write(self.profile_data, 'session_manager')



"""ThreatConnect TI Email Address"""
from ..indicator import Indicator


class EmailAddress(Indicator):
    """Unique API calls for Email Address API Endpoints"""

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties.

        Args:
            address (str): The value for this Indicator.
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): [Read-Only] The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__(
            tcex,
            sub_type='Email Address',
            api_entity='emailAddress',
            api_branch='emailAddresses',
            **kwargs
        )
        self.unique_id = kwargs.get('unique_id', kwargs.get('address'))
        self._data['address'] = self.unique_id

    def can_create(self):
        """Return True if email address can be created.

        If the address has been provided returns that the EmailAddress can be created, otherwise
        returns that the EmailAddress cannot be created.

        Returns:

        """
        return not self.data.get('address') is None

    def _set_unique_id(self, json_response):
        """Set the unique_id provided a json response.

        Args:
            json_response:
        """
        self.unique_id = json_response.get('address', '')



"""TcEx Framework Inputs module"""
# standard library
import json
import os
import sys
from argparse import Namespace

from ..utils import Utils
from .argument_parser import TcArgumentParser


class Inputs:
    """Module for handling inputs passed to App from CLI, Config, SecureParams, and AOT

    Args:
        tcex (tcex.TcEx): Instance of TcEx class.
        config (dict): A dictionary containing the configuration data for tcex and App.
        config_file (str, optional): An filename containing JSON configuration data. Defaults to
            None.
    """

    def __init__(self, tcex, config, config_file=None):
        """Initialize Class Properties.

        Input Options:

        1. All inputs from CLI args
        2. Inputs from CLI arg and from AOT params
        3. Inputs from CLI arg and from secure params
        4. All inputs from Config params
        5. Inputs from Config params and from AOT params
        6. Inputs from Config File (including encrypted file for services)
        """
        self.tcex = tcex

        # properties
        self._parsed = False
        self._parsed_resolved = False
        self.utils = Utils()

        # parser
        self.parser = TcArgumentParser(conflict_handler='resolve')  # APP-964 allow duplicate args

        # a single config file is supported, typically from and external App or service App using
        # fileParam feature
        config_file = os.getenv('TC_APP_PARAM_FILE') or config_file

        # handle config and config_file
        config_file_data = self.config_file(config_file, os.getenv('TC_APP_PARAM_KEY'))
        config.update(config_file_data)  # config_file params update config

        # create empty namespaces
        self._default_args = Namespace()
        self._default_args_resolved = Namespace()

        # update default_args namespace with config data using dict interface - for option #4 and #5
        self._default_args.__dict__.update(config)

        # parse CLI args - for option #1, #2, and #3 (additional parsing done in args() method)
        self._default_args, self._unknown_args = self.parser.parse_known_args(
            namespace=self._default_args
        )

        # update tcex default_args property for all dependent modules
        self.tcex._default_args = self._default_args

        # when running locally retrieve any args from the results.tc file
        self._results_tc_args()

        # register token as soon as possible (required for the any API call including secure params)
        self.register_token()

        # load aot params from redis (blocking) - for options #2 and #5
        self._load_aot_params()

        # load secure params from API - for options #3
        self._load_secure_params()

        # add default args namespace to parser for add_argument() method
        # used to covert any required args in Apps to default values from namespace
        self.parser.namespace = self._default_args

        # update logging now that all required tcex logging parameters are loaded
        self.update_logging()

    def _get_secure_params(self):
        """Load secure params from the API.

        # API Response:

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1

            {
                "inputs":
                    {
                        "tc_playbook_db_type": "Redis",
                        "fail_on_error": true,
                        "api_default_org": "TCI"
                    }
            }

        Returns:
            dict: Parameters ("inputs") from the TC API.
        """
        self.tcex.log.info('feature=inputs, event=loading-secure-params')

        # Retrieve secure params from API
        r = self.tcex.session.get('/internal/job/execution/parameters')

        # check for bad status code and response that is not JSON
        if not r.ok:
            err = r.text or r.reason
            raise RuntimeError(f'Error retrieving secure params from API ({err}).')

        secure_params = {}
        try:
            secure_params = r.json()['inputs']
        except (AttributeError, KeyError, TypeError, ValueError):  # pragma: no cover
            err = r.text or r.reason
            raise RuntimeError(f'Error retrieving secure params from API ({err}).')

        return secure_params

    def _load_aot_params(self):
        """Block and retrieve params from Redis."""
        if self._default_args.tc_aot_enabled is True:
            # update default_args with AOT params
            params = self.aot_blpop()
            updated_params = self.update_params(params)
            # log number of params returned from AOT
            self.tcex.log.info(
                f'feature=inputs, event=loaded-aot-params, count={len(updated_params)}'
            )
            self.config(updated_params)

    def _load_secure_params(self):
        """Parse args and return default args."""
        if self._default_args.tc_secure_params is True:
            # update default_args with secure params from API
            params = self._get_secure_params()
            updated_params = self.update_params(params)
            self.tcex.log.info(
                f'feature=inputs, event=loaded-secure-params, count={len(updated_params)}'
            )
            self.config(updated_params)

    def _results_tc_args(self):  # pragma: no cover
        """Read data from results_tc file from previous run of app.

        This method is only required when not running from within the TcEX platform and is only
        intended for testing apps locally.
        """
        results = []
        if os.access(self.default_args.tc_out_path, os.W_OK):
            result_file = f'{self.default_args.tc_out_path}/results.tc'
        else:
            result_file = 'results.tc'
        if os.path.isfile(result_file):
            with open(result_file) as rh:
                results = rh.read().strip().split('\n')
            os.remove(result_file)
        for line in results:
            if not line or ' = ' not in line:
                continue
            key, value = line.split(' = ')
            if value == 'true':
                value = True
            elif value == 'false':
                value = False
            elif not value:
                value = None
            setattr(self._default_args, key, value)

    def aot_blpop(self):
        """Subscribe to AOT action channel."""
        res = None
        if self._default_args.tc_playbook_db_type == 'Redis':
            try:
                self.tcex.log.info('feature=inputs, event=blocking-for-aot')
                msg_data = self.tcex.redis_client.blpop(
                    keys=self._default_args.tc_action_channel,
                    timeout=self._default_args.tc_terminate_seconds,
                )

                if msg_data is None:
                    self.tcex.exit(0, 'AOT subscription timeout reached.')

                msg_data = json.loads(msg_data[1])
                msg_type = msg_data.get('type', 'terminate')
                if msg_type == 'execute':
                    res = msg_data.get('params', {})
                elif msg_type == 'terminate':
                    self.tcex.exit(0, 'Received AOT terminate message.')
                else:  # pragma: no cover
                    self.tcex.log.warn(
                        f'feature=inputs, event=unsupported-aot-message, type={msg_type}'
                    )
                    res = self.aot_blpop()
            except Exception as e:  # pragma: no cover
                self.tcex.exit(1, f'Exception during AOT subscription ({e}).')

        return res

    def args(self, parse=False):
        """Parse args if they have not already been parsed and return the Namespace for args.

        .. Note:: Accessing args should only be done directly in the App.

        Returns:
            (namespace): ArgParser parsed arguments.
        """
        if not self._parsed or parse:
            # initialize default args
            args, self._unknown_args = self.parser.parse_known_args(namespace=self._default_args)
            self.config(args.__dict__, False)

            # special case for service Apps
            if (
                not self._default_args.tc_svc_id
                and self._default_args.tc_svc_client_topic is not None
            ):  # pragma: no cover
                # get the service id as third part of the service
                # --tc_svc_client_topic svc-client-cc66d36344787779ccaa8dbb5e09a7ab
                setattr(
                    self._default_args,
                    'tc_svc_id',
                    self._default_args.tc_svc_client_topic.split('-')[2],
                )

            # special case for service Apps
            if not self._default_args.tc_job_id and self._default_args.tc_token is not None:
                # get the job id from the token
                setattr(
                    self._default_args, 'tc_job_id', self._default_args.tc_token.split(':')[-2],
                )

            # set parsed bool to ensure args are only parsed once
            self._parsed = True

            # log unknown arguments
            self.unknown_args()

        return self._default_args

    def config(self, config_data, preserve=True):
        """Add configuration data to update default_args.

        Below are the default args that the TcEx frameworks supports. Any App specific args
        should be included in the provided data.

        .. code-block:: javascript

            {
              "api_access_id": "$env.API_ACCESS_ID",
              "api_default_org": "$env.API_DEFAULT_ORG",
              "api_secret_key": "$envs.API_SECRET_KEY",
              "tc_api_path": "$env.TC_API_PATH",
              "tc_log_level": "debug",
              "tc_log_path": "log",
              "tc_owner": "MyOwner",
              "tc_proxy_host": "$env.TC_PROXY_HOST",
              "tc_proxy_password": "$envs.TC_PROXY_PASSWORD",
              "tc_proxy_port": "$env.TC_PROXY_PORT",
              "tc_proxy_tc": false,
              "tc_proxy_username": "$env.TC_PROXY_USERNAME"
            }

        Args:
            config_data (dict): A dictionary of configuration values.
            preserve (bool): Don't overwrite arg values define in sys.argv
        """
        if isinstance(config_data, dict):
            if preserve:
                # on env server core doesn't send all required values on cli. inputs that
                # come in via secureParams needs to be updated, but not all of them (e.g. log_path).
                # this code will only update new inputs that are not provided via sys argv.
                for key in list(config_data):
                    if f'--{key}' in sys.argv:
                        del config_data[key]

            # update the arg Namespace via dict
            self._default_args.__dict__.update(config_data)

            # register token as soon as possible
            self.register_token()

    def config_file(self, filename, key=None):
        """Load configuration data from provided file and update default_args.

        Args:
            config (str): The configuration file name.
            key (str): The configuration file encryption key.

        Returns:
            dict: The JSON contents of the file as a dict.
        """
        file_content = {}
        if filename is not None and os.path.isfile(filename):
            if key is not None:
                try:
                    # read encrypted file from "in" directory
                    with open(filename, 'rb') as fh:
                        encrypted_contents = fh.read()

                    result = self.utils.decrypt_aes_cbc(key, encrypted_contents)
                    file_content = json.loads(result.decode('utf-8'))
                    file_content = self.update_params(file_content)

                    # delete file
                    os.unlink(filename)
                except Exception:  # pragma: no cover
                    self.tcex.log.error(
                        f'feature=inputs, event=config-decryption-failure, filename={filename}'
                    )
                self.tcex.log.info(
                    f'feature=inputs, event=config-loaded, count={len(file_content)}'
                )
            else:
                try:
                    with open(filename) as fh:
                        file_content = json.load(fh)
                except ValueError:  # pragma: no cover
                    self.tcex.log.error(
                        f'feature=inputs, event=config-parse-failure, filename={filename}'
                    )
        elif filename is not None:  # pragma: no cover
            self.tcex.log.error(f'feature=inputs, event=config-load-failure, filename={filename}')
        return file_content

    @property
    def default_args(self):
        """Parse args and return default args."""
        return self._default_args

    @property
    def params(self):
        """Return input params."""
        # return self._default_args.__dict__
        self.args()  # ensure all args are parsed
        return self._default_args

    def register_token(self):
        """Register token if provided in args (non-service Apps)"""
        # TODO: swap MainThread with threading.current_thread().name ?
        if self._default_args.tc_token is not None:
            self.tcex.token.register_token(
                'MainThread', self._default_args.tc_token, self._default_args.tc_token_expires
            )

    def resolved_args(self, parse=False):
        """Return namespace of args that have all playbook variables automatically resolved.

        .. Note:: Accessing resolved_args should only be done directly in the App.

        Returns:
            (namespace): ArgParser parsed arguments with Playbook variables automatically resolved.
        """
        if not self._parsed_resolved or parse:  # only resolve once
            self.args()

            # create new args Namespace for resolved args
            self._default_args_resolved = Namespace()

            # iterate over args and resolve any playbook variables
            for arg in vars(self._default_args):
                arg_val = getattr(self._default_args, arg)
                if arg not in self.tc_reserved_args:
                    if isinstance(arg_val, (str)):
                        arg_val = self.tcex.playbook.read(arg_val)
                setattr(self._default_args_resolved, arg, arg_val)

            # set parsed bool to ensure args are only parsed once
            self._parsed_resolved = True
        return self._default_args_resolved

    @property
    def resolved_params(self):
        """Return input params."""
        # return self._default_args_resolved.__dict__
        return self._default_args_resolved

    @property
    def tc_bool_args(self):
        """Return a list of default ThreatConnect Args that are booleans."""
        return [
            'apply_proxy_external',
            'apply_proxy_ext',
            'apply_proxy_tc',
            'batch_halt_on_error',
            'tc_aot_enabled',
            'tc_log_to_api',
            'tc_proxy_external',
            'tc_proxy_tc',
            'tc_secure_params',
            'tc_verify',
        ]

    @property
    def tc_reserved_args(self):
        """Return a list of *all* ThreatConnect reserved arg values."""
        return [
            'api_access_id',
            'api_default_org',
            'api_secret_key',
            'batch_action',
            'batch_chunk',
            'batch_halt_on_error',
            'batch_interval_max',
            'batch_poll_interval',
            'batch_write_type',
            'logging',
            'tc_api_path',
            'tc_in_path',
            'tc_log_file',
            'tc_log_level',
            'tc_log_path',
            'tc_log_to_api',
            'tc_out_path',
            'tc_playbook_db_context',
            'tc_playbook_db_path',
            'tc_playbook_db_port',
            'tc_playbook_db_type',
            'tc_playbook_out_variables',
            'tc_proxy_host',
            'tc_proxy_port',
            'tc_proxy_username',
            'tc_proxy_password',
            'tc_proxy_external',
            'tc_proxy_tc',
            'tc_secure_params',
            'tc_temp_path',
            'tc_token',
            'tc_token_expires',
            'tc_user_id',
        ]

    def unknown_args(self):
        """Log argparser unknown arguments.

        This method is only applicable for args passed on the CLI.

        Args:
            args (list): List of unknown arguments
        """
        for u in self._unknown_args:
            if u == 'run':
                continue
            self.tcex.log.warning(f'feature=inputs, event=unsupported-arg, arg={u}')

    def update_logging(self):
        """Update the TcEx logger with appropriate handlers."""
        # custom loggers passed into tcex would not have log_info method
        if not hasattr(self.tcex.logger, 'log_info'):
            return

        if self._default_args.tc_log_level is None:
            # some Apps use logging while other us tc_log_level. ensure tc_log_level is always
            # available.
            self._default_args.tc_log_level = self._default_args.logging

        self.tcex.logger.log_info(self._default_args)

        # add api handler
        if self._default_args.tc_token is not None and self._default_args.tc_log_to_api:
            self.tcex.logger.add_api_handler(level=self.tcex.default_args.tc_log_level)

        # add rotating log handler
        self.tcex.logger.add_rotating_file_handler(
            name='rfh',
            filename=self._default_args.tc_log_file,
            path=self._default_args.tc_log_path,
            backup_count=self._default_args.tc_log_backup_count,
            max_bytes=self._default_args.tc_log_max_bytes,
            level=self.tcex.default_args.tc_log_level,
        )

        # replay cached log events
        self.tcex.logger.replay_cached_events(handler_name='cache')

    def update_params(self, params):
        """Update params provided by AOT and Secure Params to be of the proper value and type.

        Args:
            params (dict): A dictionary containing params to update default_args
        """
        updated_params = {}
        for arg, value in params.items():
            # ThreatConnect secure/AOT params could be updated in the future to proper JSON format.
            # MultiChoice data should be represented as JSON array and Boolean values should be a
            # JSON boolean and not a string.
            param_data = self.tcex.ij.params_dict.get(arg) or {}
            param_type = param_data.get('type', '').lower()
            param_allow_multiple = self.utils.to_bool(param_data.get('allowMultiple', False))

            if param_type == 'multichoice' or param_allow_multiple:
                # update delimited value to an array for params that have type of MultiChoice.
                if value is not None and not isinstance(value, dict):
                    value = value.split(self.tcex.ij.list_delimiter)
            elif param_type == 'boolean':
                # convert boolean input that are passed in as a string ("true" -> True)
                value = self.utils.to_bool(value)
            elif arg in self.tc_bool_args:
                # convert default boolean args that are passed in as a string ("true" -> True)
                value = self.utils.to_bool(value)

            # add args and updated value to dict
            updated_params[arg] = value

        # update args
        return updated_params



"""ThreatConnect STIX module"""
# standard library
from typing import Union

# third-party
from stix2 import WindowsRegistryKey

# first-party
from tcex.stix.model import StixModel  # pylint: disable=cyclic-import


class StixRegistryKeyObject(StixModel):
    """STIX Threat Actor object."""

    def consume(self, stix_data: Union[list, dict]):
        """Convert STIX Windows Registry Key Cyber Observables to ThreatConnect Indicators.

        Args:
            stix_data: STIX Windows Registry Key Cyber Observables.

        Yields:
            ThreatConnect registry key indicators.
        """
        if isinstance(stix_data, dict):
            stix_data = [stix_data]

        for data in stix_data:
            mapper = {
                'type': 'Registry Key',
                'Key Name': '@.key',
                'xid': '@.id',
                'attributes': [{'type': 'External ID', 'value': '@.id'}],
            }
            if not data.get('values'):
                yield from self._map(data, mapper)
            else:
                for i in range(len(data.get('values'))):
                    mapper['Value Name'] = f'@.values[{i}].name'
                    mapper['Value Type'] = f'@.values[{i}].data_type'
                    mapper['attributes'].append(
                        {'type': 'Value Data', 'value': f'@.values[{i}].data'}
                    )
                    yield from self._map(data, mapper)

    def produce(self, tc_data: Union[list, dict], **kwargs):
        """Produce STIX 2.0 JSON object from TC API response.

        .. code:: json

            {
              "type": "windows-registry-key",
              "spec_version": "2.1",
              "id": "windows-registry-key--2ba37ae7-2745-5082-9dfd-9486dad41016",
              "key": "hkey_local_machine\\system\\bar\\foo",
              "values": [
                {
                  "name": "Foo",
                  "data": "qwerty",
                  "data_type": "REG_SZ"
                },
                {
                  "name": "Bar",
                  "data": "42",
                  "data_type": "REG_DWORD"
                }
              ]
            }
        """
        if isinstance(tc_data, list) and len(tc_data) > 0 and 'summary' in tc_data[0]:
            for data in tc_data:
                summary = data.get('summary').split(':')
                data['Key Name'] = summary[0].trim()
                data['Value Name'] = summary[1].trim()
                data['Value Type'] = summary[2].trim()
                del data['summary']

        mapper = {
            'id': '@.id',
            'key': '@.Key Name',
            'value': [{'name': '@.Value Name', 'data': '', 'data_type': '@.Value Type'}],
            'spec_version': '2.1',
            'type': 'windows-registry-key',
        }

        yield from (WindowsRegistryKey(**stix_data) for stix_data in self._map(tc_data, mapper))



"""Several helper functions that produce a STIX pattern for a threatconnect indicator."""
# standard library
import ipaddress


def address_stix_pattern_producer(data):
    """Convert an address from TC to a STIX pattern."""
    if isinstance(ipaddress.ip_address(data.get('summary')), ipaddress.IPv6Address):
        return f"[ipv6-addr:value = '{data.get('summary')}']"

    return f"[ipv4-addr:value = '{data.get('summary')}']"


def cidr_stix_pattern_producer(data):
    """Convert a CIDR from TC to a STIX pattern."""
    if isinstance(ipaddress.ip_network(data.get('summary'), strict=False), ipaddress.IPv6Network):
        return f"[ipv6-addr:value = '{data.get('summary')}']"

    return f"[ipv4-addr:value = '{data.get('summary')}']"


def file_stix_pattern_producer(data):
    """Convert a File from TC to a STIX pattern."""
    expressions = []
    for _hash in data.get('summary', '').split(' : '):
        if len(_hash) == 32:
            expressions.append(f"file:hashes.md5 = '{_hash}'")
        elif len(_hash) == 40:
            expressions.append(f"file:hashes.sha1 = '{_hash}'")
        elif len(_hash) == 64:
            expressions.append(f"file:hashes.sha256 = '{_hash}'")

    return f'[{" OR ".join(expressions)}]'


def asn_stix_pattern_producer(data):
    """Convert an ASN from TC to a STIX pattern."""
    return f"[autonomous-system:name = '{data.get('summary')}']"


def email_address_stix_pattern_producer(data):
    """Convert an email address from TC to a STIX pattern."""
    return f"[email-addr:value = '{data.get('summary')}']"


def url_stix_pattern_producer(data):
    """Convert a URL from TC to a STIX pattern."""
    return f"[url:value = '{data.get('summary')}']"


def host_stix_pattern_producer(data):
    """Convert a host from TC to a STIX pattern."""
    return f"[domain-name:value = '{data.get('summary')}']"


def registery_key_stix_pattern_producer(data):
    """Convert a Registry Key from TC to a STIX pattern."""
    key_name, value_name, value_type = data.get('summary').split(':')

    parts = []

    if key_name.strip():
        parts.append(f"windows-registry-key:key = '{key_name.strip()}'")

    if value_name.strip():
        parts.append(f"windows-registry-key:values.name = '{value_name.strip()}'")

    if value_type.strip():
        parts.append(f"windows-registry-key:values.data_type = '{value_type.strip()}'")

    return f'[{" AND ".join(parts)}]'



"""TcEx Framework Module for working with DataStore in the ThreatConnect Platform."""
# standard library
from typing import Optional

# third-party
from requests.models import Response


class DataStore:
    """TcEx DataStore Class.

    **domain**

    This can be either “system”, “organization”, or “local”. Any string not defined as these
    3 will return a “Resource Not Found” error response.

    **dataType**

    This is a free form index type name as allowed by ElasticSearch. The ThreatConnect API will
    use this resource verbatim for ElasticSearch calls.

    **searchCommand**

    The URI string after the typeName is assumed to be ElasticSearch relevant information and is
    passed directly to ElasticSearch.

    Args:
        tcex (TcEx): An instance of TcEx.
        domain (str): A value of “system”, “organization”, or “local”.
        data_type (str): A free form type name for the data.
        mapping (Optional[dict] = None): Defaults to None. Elasticsearch mappings data.
    """

    def __init__(self, tcex: object, domain: str, data_type: str, mapping: Optional[dict] = None):
        """Initialize class properties."""
        self.tcex: object = tcex
        self.domain: str = domain
        self.data_type: str = data_type
        self.mapping: Optional[dict] = mapping or {'dynamic': False}

        # setup
        self._create_index()  # create the initial index.
        self._update_mappings()  # update mappings

    def _create_index(self):
        """Create index if it doesn't exist."""
        if not self.index_exists:
            rid = 'temp-index-create'
            index_settings = {}
            headers = {'Content-Type': 'application/json', 'DB-Method': 'POST'}
            url = f'/v2/exchange/db/{self.domain}/{self.data_type}/{rid}'
            r: Response = self.tcex.session.post(url, json=index_settings, headers=headers)

            if not r.ok:
                error: str = r.text or r.reason
                self.tcex.handle_error(800, [r.status_code, error])
            self.tcex.log.debug(
                f'creating index. status_code: {r.status_code}, response: "{r.text}".'
            )
            # delete temporary record
            self.delete(rid, False)

    def _token_available(self):
        """Raise an error if token is not available.

        This check needs to be made after args are parsed to ensure token module has initialized.
        """
        # This module requires a token.
        if self.tcex.token.token is None:  # pragma: no cover
            raise RuntimeError(
                'The DataModel TcEx Module requires a Token to '
                'interact with the ThreatConnect platform.'
            )

    def _update_mappings(self):
        """Update the mappings for the current index."""
        headers = {'Content-Type': 'application/json', 'DB-Method': 'PUT'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/_mappings'
        r: Response = self.tcex.session.post(url, json=self.mapping, headers=headers)
        self.tcex.log.debug(f'update mapping. status_code: {r.status_code}, response: "{r.text}".')

    @property
    def index_exists(self) -> bool:
        """Check to see if index exists."""
        headers = {'Content-Type': 'application/json', 'DB-Method': 'GET'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/_search'
        r: Response = self.tcex.session.post(url, headers=headers)
        if not r.ok:
            self.tcex.log.warning(f'The provided index was not found ({r.text}).')
            return False
        return True

    def add(self, rid: str, data: dict, raise_on_error: Optional[bool] = True) -> dict:
        """Write data to the DataStore. Alias for post() method.

        Args:
            rid (str): The record identifier.
            data (dict): The record data.
            raise_on_error (Optional[bool] = True): If True and not r.ok this
                method will raise a RunTimeError.

        Returns:
            dict : The response data.
        """
        return self.post(rid, data, raise_on_error)

    def delete(self, rid: str, raise_on_error: Optional[bool] = True) -> dict:
        """Delete a record from the index using provide Id.

        **Example Response**

        .. code-block:: json

            {
              "_index": "$local.usr5_pytest",
              "_type": "pytest",
              "_id": "three",
              "_version": 4,
              "result": "deleted",
              "_shards": {
                "total": 2,
                "successful": 1,
                "failed": 0
              },
              "_seq_no": 45,
              "_primary_term": 1
            }

        Args:
            rid (str): The record identifier.
            raise_on_error (Optional[bool] = True): If True and not r.ok this
                method will raise a RunTimeError.

        Returns:
            dict : The response data.
        """
        self._token_available()  # ensure a token is available

        response_data = None
        headers = {'Content-Type': 'application/json', 'DB-Method': 'DELETE'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/{rid}'
        r: Response = self.tcex.session.post(url, headers=headers)
        self.tcex.log.debug(f'datastore delete status code: {r.status_code}')
        if r.ok and 'application/json' in r.headers.get('content-type', ''):
            response_data: dict = r.json()
        else:
            error: str = r.text or r.reason
            self.tcex.handle_error(805, ['delete', r.status_code, error], raise_on_error)
        return response_data

    def get(
        self,
        rid: Optional[str] = None,
        data: Optional[dict] = None,
        raise_on_error: Optional[bool] = True,
    ) -> dict:
        """Get data from the DataStore.

        **Example Response**

        .. code-block:: json

            {
              "_index": "$local.usr5_pytest",
              "_type": "pytest",
              "_id": "two",
              "_version": 1,
              "found": true,
              "_source": {
                "two": 2
              }
            }

        Args:
            rid (Optional[str] = None): The record identifier.
            data (Optional[dict] = None): A search query
            raise_on_error (Optional[bool] = True): If True and not r.ok this
                method will raise a RunTimeError.

        Returns:
            dict : Python request response.
        """
        self._token_available()  # ensure a token is available

        response_data = None
        headers = {'Content-Type': 'application/json', 'DB-Method': 'GET'}
        if rid is None:
            url = f'/v2/exchange/db/{self.domain}/{self.data_type}/'
        else:
            url = f'/v2/exchange/db/{self.domain}/{self.data_type}/{rid}'
        r: Response = self.tcex.session.post(url, json=data, headers=headers)
        self.tcex.log.debug(f'datastore get status code: {r.status_code}')
        if 'application/json' in r.headers.get('content-type', ''):
            # as long as the content is JSON set the value
            try:
                response_data: dict = r.json()
            except Exception as e:  # pragma: no cover
                # This issue should be addressed by core in a future release.
                self.tcex.log.warning(
                    'DataStore API returned a non-JSON response, even though '
                    f'content-type was application/json: {r.content} error: {e}'
                )
        if not r.ok:
            error: str = r.text or r.reason
            self.tcex.handle_error(805, ['get', r.status_code, error], raise_on_error)
        return response_data

    def post(self, rid: str, data: str, raise_on_error: Optional[bool] = True) -> dict:
        """Write data to the DataStore.

        **Example Response**

        .. code-block:: json

            {
              "_index": "$local.usr5_pytest",
              "_type": "pytest",
              "_id": "one",
              "_version": 31,
              "result": "updated",
              "_shards": {
                "total": 2,
                "successful": 1,
                "failed": 0
              },
              "_seq_no": 53,
              "_primary_term": 1
            }

        Args:
            rid (str): The record identifier. If a value of None is passed an identifier elastic
                search will automatically generate a id.
            data (dict): The record data.
            raise_on_error (Optional[bool] = True): If True and not r.ok this
                method will raise a RunTimeError.

        Returns:
            dict : The response data.
        """
        self._token_available()  # ensure a token is available

        response_data = None
        headers = {'Content-Type': 'application/json', 'DB-Method': 'POST'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/'
        if rid is not None:
            url = f'{url}{rid}'

        r: Response = self.tcex.session.post(url, json=data, headers=headers)
        self.tcex.log.debug(f'datastore post status code: {r.status_code}')

        if r.ok and 'application/json' in r.headers.get('content-type', ''):
            response_data: dict = r.json()
        else:
            error: str = r.text or r.reason
            self.tcex.handle_error(805, ['post', r.status_code, error], raise_on_error)
        return response_data

    def put(self, rid: str, data: dict, raise_on_error: Optional[bool] = True) -> dict:
        """Update the data for the provided Id.

        **Example Response**

        .. code-block:: json

            {
              "_index": "$local.usr5_pytest",
              "_type": "pytest",
              "_id": "one",
              "_version": 33,
              "result": "updated",
              "_shards": {
                "total": 2,
                "successful": 1,
                "failed": 0
              },
              "_seq_no": 55,
              "_primary_term": 1
            }

        Args:
            rid (str): The record identifier.
            data (dict): A search query
            raise_on_error (Optional[bool] = True): If True and not r.ok this
                method will raise a RunTimeError.

        Returns:
            dict : The response dict.
        """
        self._token_available()  # ensure a token is available

        response_data = None
        headers = {'Content-Type': 'application/json', 'DB-Method': 'PUT'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/{rid}'

        r: Response = self.tcex.session.post(url, json=data, headers=headers)
        self.tcex.log.debug(f'datastore put status code: {r.status_code}')

        if r.ok and 'application/json' in r.headers.get('content-type', ''):
            response_data: dict = r.json()
        else:
            error: str = r.text or r.reason
            self.tcex.handle_error(805, ['put', r.status_code, error], raise_on_error)
        return response_data

    def update(self, rid: str, data: dict, raise_on_error: Optional[bool] = True):
        """Update the for the provided Id. Alias for put() method.

        Args:
            rid (str): The record identifier.
            data (dict): The record data.
            raise_on_error (Optional[bool] = True): If True and not r.ok this
                method will raise a RunTimeError.

        Returns:
            dict : The response data.
        """
        return self.put(rid, data, raise_on_error)



"""Testing module for TcEx Framework"""
# flake8: noqa
# first-party
from tcex.testing.monkeypatch import monkeypatch, register_monkeypatch
from tcex.testing.stage_data import Stager
from tcex.testing.test_case_api_service import TestCaseApiService
from tcex.testing.test_case_job import TestCaseJob
from tcex.testing.test_case_playbook import TestCasePlaybook
from tcex.testing.test_case_trigger_service import TestCaseTriggerService
from tcex.testing.test_case_webhook_trigger_service import TestCaseWebhookTriggerService
from tcex.testing.validate_data import Validator



"""ThreatConnect Batch Import Module"""
# standard library
import json
from typing import Callable, Optional


class Attribute:
    """ThreatConnect Batch Attribute Object"""

    __slots__ = ['_attribute_data', '_valid']

    def __init__(
        self,
        attr_type: str,
        attr_value: str,
        displayed: Optional[bool] = False,
        source: Optional[str] = None,
        formatter: Optional[Callable[[str], str]] = None,
    ) -> None:
        """Initialize Class Properties.

        Args:
            attr_type: The ThreatConnect defined attribute type.
            attr_value: The value for this attribute.
            displayed: If True the supported attribute will be marked for display.
            source: The source value for this attribute.
            formatter: A callable that take a single attribute
                value and return a single formatted value.
        """
        self._attribute_data = {'type': attr_type}
        if displayed:
            self._attribute_data['displayed'] = displayed
        # format the value
        if formatter is not None:
            attr_value = formatter(attr_value)
        self._attribute_data['value'] = attr_value
        # add source if provided
        if source is not None:
            self._attribute_data['source'] = source
        # is attr_value not null or ''
        self._valid = True
        # check for None and '' value only.
        if attr_value in [None, '']:
            self._valid = False

    @property
    def data(self) -> dict:
        """Return Attribute data."""
        return self._attribute_data

    @property
    def displayed(self) -> bool:
        """Return Attribute displayed."""
        return self._attribute_data.get('displayed')

    @displayed.setter
    def displayed(self, displayed: bool):
        """Set Attribute displayed."""
        self._attribute_data['displayed'] = displayed

    @property
    def source(self) -> str:
        """Return Attribute source."""
        return self._attribute_data.get('source')

    @source.setter
    def source(self, source: str):
        """Set Attribute source."""
        self._attribute_data['source'] = source

    @property
    def type(self) -> str:
        """Return attribute value."""
        return self._attribute_data.get('type')

    @property
    def valid(self) -> bool:
        """Return valid value."""
        return self._valid

    @property
    def value(self) -> str:
        """Return attribute value."""
        return self._attribute_data.get('value')

    def __str__(self) -> str:
        """Return string represtentation of object."""
        return json.dumps(self.data, indent=4)



#!/usr/bin/env python
"""TcEx Framework Template Download/Generation Module."""
# standard library
import os
import sys
from typing import Any, Optional

# third-party
import colorama as c
import requests

try:
    # third-party
    from mako.template import Template
except ImportError:
    # mako is only required for local development
    pass

from ..utils import Utils

# autoreset colorama
c.init(autoreset=True, strip=False)


class TemplateBase:
    """Base class for template module.

    Args:
        branch (str): The git branch of tcex to use for downloads.
        profile (Profile): An instance of Profile.
    """

    def __init__(self, profile: object, branch: Optional[str] = 'master'):
        """Initialize class properties."""
        self.url = f'https://raw.githubusercontent.com/ThreatConnect-Inc/tcex/{branch}/app_init'
        self.profile = profile

        # properties
        self._download_results = []
        self._template_render_results = []

    @staticmethod
    def _short_string(
        string: str, length: Optional[int] = 65, prepend_ellipsis: Optional[bool] = True
    ) -> str:
        """Return a shortened destination."""
        ellipsis = ''
        if prepend_ellipsis and len(string) > length:
            ellipsis = '...'
            length -= len(ellipsis)
        return f'{ellipsis}{string[-length:]}'

    @staticmethod
    def download_template(url: str) -> Any:
        """Return template file"""
        r: object = requests.get(url, allow_redirects=True, headers={'Cache-Control': 'no-cache'})
        if not r.ok:
            raise RuntimeError(f'Could not download template file ({url}).')
        return r.content

    def print_results(self, results: list) -> None:
        """Print the results."""
        if results:
            # get the max length of url
            destination_length = 0
            destination_max = 50
            url_length = 0
            for r in results:
                if len(r.get('destination')) > destination_length:
                    destination_length = min(len(r.get('destination')) + 5, destination_max)
                if len(r.get('url')) > url_length:
                    url_length = len(r.get('url')) + 5

            # TODO: fix this - overriding url_length temporarily
            url_length = 120
            destination_length = 50

            # title row
            print(
                f"{c.Style.BRIGHT}{c.Fore.CYAN}{'Download URL:'!s:<{url_length}}"
                f"{'Destination:'!s:<{destination_length}}"
                f"{'Status:'!s:<10}"
            )
            for r in results:
                destination = self._short_string(
                    r.get('destination'), length=(destination_length - 5)
                )
                status = r.get('status')
                url = r.get('url')

                # get status color
                status_color = f'{c.Style.BRIGHT}{c.Fore.RED}'
                if status == 'Success':
                    status_color = c.Fore.GREEN
                elif status == 'Skipped':
                    status_color = f'{c.Style.BRIGHT}{c.Fore.YELLOW}'

                # data row
                print(
                    f'{url!s:<{url_length}}'
                    f'{destination!s:<{destination_length}}'
                    f'{status_color}{status!s:<10}'
                )

    def print_download_results(self) -> None:
        """Print the download results."""
        self.print_results(self._download_results)

    def print_template_render_results(self) -> None:
        """Print the download results."""
        self.print_results(self._template_render_results)

    def render(
        self,
        template_url: str,
        destination: str,
        variables: dict,
        overwrite: Optional[bool] = False,
    ):
        """Render the provided template"""
        if 'mako' not in sys.modules:
            print(
                f'{c.Fore.RED}Missing mako module. Try installing "pip install tcex[development]"'
            )
            sys.exit(1)

        status = 'Failed'
        if not os.path.isfile(destination) or overwrite:
            template_data: str = self.download_template(template_url)
            template: object = Template(template_data)  # nosec
            rendered_template: str = template.render(**variables)
            with open(destination, 'w') as f:
                f.write(rendered_template)
            status = 'Success'
        else:
            status = 'Skipped'

        self._template_render_results.append(
            {'destination': destination, 'status': status, 'url': template_url}
        )


class CustomTemplates(TemplateBase):
    """Custom method Template Class

    Args:
        branch (str): The git branch of tcex to use for downloads.
        profile (Profile): An instance of Profile.
    """

    def __init__(self, profile: Optional[object] = None, branch: Optional[str] = 'master'):
        """Initialize Class properties."""
        super().__init__(profile, branch)

    def custom_py(self):
        """Render the custom.py template."""
        filename: str = os.path.join(self.profile.test_directory, 'custom.py')
        variables = {'runtime_level': self.profile.ij.runtime_level.lower()}
        if not os.path.isfile(filename):
            self.render(f'{self.url}/tests/custom.py.tpl', filename, variables)

    def custom_feature_py(self):
        """Render the custom.py template."""
        filename = os.path.join(self.profile.feature_directory, 'custom_feature.py')
        variables = {'runtime_level': self.profile.ij.runtime_level.lower()}
        if not os.path.isfile(filename):
            self.render(f'{self.url}/tests/custom_feature.py.tpl', filename, variables)

    def render_templates(self):
        """Render the templates and write to disk conditionally."""
        self.custom_py()
        self.custom_feature_py()


class DownloadTemplates(TemplateBase):
    """Template Download Module.

    Args:
        branch (str): The git branch of tcex to use for downloads.
        profile (Profile): An instance of Profile.
    """

    def __init__(self, profile: Optional[object] = None, branch: Optional[str] = 'master'):
        """Initialize Class properties."""
        super().__init__(profile, branch)

        # properties
        self.utils = Utils()

    def download_file(
        self,
        url: str,
        destination: str,
        overwrite: Optional[bool] = False,
        default_choice: Optional[str] = 'yes',
    ):
        """Download the provided source file to the provided destination.

        Args:
            url (str): The full URL where the file can be downloaded.
            destination (str): The fully qualified path to the destination file.
            overwrite (bool, str): If True the file will be overwritten, if value
                is "prompt" the user will get prompted to choose.
            default_choice (str): The value for the default choice on the download prompt.
        """

        prompted = False
        if (
            not isinstance(overwrite, bool)
            and overwrite.lower() == 'prompt'
            and os.path.isfile(destination)
        ):
            overwrite: bool = self.download_prompt(destination, default_choice)
            prompted = True

        status = 'Failed'
        if not os.path.isfile(destination) or overwrite is True:
            r = requests.get(url, allow_redirects=True, headers={'Cache-Control': 'no-cache'})
            if not r.ok:
                raise RuntimeError(f'Could not download template file ({url}).')

            # write destination file
            open(destination, 'wb').write(r.content)
            status = 'Success'
        elif prompted:
            status = 'Skipped'
        else:
            print(f'{c.Fore.RED}Destination file {destination} already exists.')
            status = 'Skipped'

        self._download_results.append({'destination': destination, 'status': status, 'url': url})

    def download_prompt(self, destination: str, default_choice: str):
        """Present the download prompt for the user.

        Args:
            destination (str): The fully qualified path to the destination file.
            default_choice (str): The value for the default choice on the download prompt.

        Returns:
            bool: True if the user select yes.
        """
        option_text = self.download_prompt_option_text(default_choice)

        print(f'''{c.Style.BRIGHT}{'File:'!s:<15}{c.Fore.CYAN}{self._short_string(destination)}''')
        message = (
            f'{c.Fore.MAGENTA}Replace '
            f'{c.Style.BRIGHT}{c.Fore.WHITE}({option_text}):{c.Fore.RESET} '
        )
        response: str = input(message).strip()  # nosec
        if not response:
            response = default_choice
        return self.utils.to_bool(response)

    @staticmethod
    def download_prompt_option_text(default_choice: str):
        """Return the default value and option text.

        Args:
            default_choice: The default choice for the prompt.

        Returns:
            str: The options text string.
        """
        valid_values = ['yes', 'no']

        options = []
        for v in valid_values:
            o = v
            if v.lower() == default_choice.lower():
                o = f'[{v}]'
            options.append(o)
        return f'''{'/'.join(options)}'''

    def init_common_files(self, template: Optional[str] = None):
        """Download common files."""
        self.download_file(
            f'{self.url}/.pre-commit-config.yaml',
            destination='.pre-commit-config.yaml',
            overwrite=True,
        )
        self.download_file(
            url=f'{self.url}/coveragerc',
            destination='.coveragerc',
            overwrite='prompt',
            default_choice='yes',
        )
        self.download_file(
            url=f'{self.url}/gitignore',
            destination='.gitignore',
            overwrite='prompt',
            default_choice='yes',
        )
        self.download_file(
            url=f'{self.url}/pyproject.toml', destination='pyproject.toml', overwrite=True
        )
        self.download_file(
            url=f'{self.url}/README.md',
            destination='README.md',
            overwrite='prompt',
            default_choice='no',
        )
        self.download_file(
            url=f'{self.url}/requirements.txt',
            destination='requirements.txt',
            overwrite='prompt',
            default_choice='no',
        )
        self.download_file(url=f'{self.url}/setup.cfg', destination='setup.cfg', overwrite=True)
        self.download_file(
            f'{self.url}/{template}/app.py',
            destination='app.py',
            overwrite='prompt',
            default_choice='no',
        )
        self.download_file(f'{self.url}/app_lib.py', destination='app_lib.py', overwrite=True)
        if template and not template.startswith('external'):
            self.download_file(
                f'{self.url}/{template}/args.py',
                destination='args.py',
                overwrite='prompt',
                default_choice='no',
            )
            self.download_file(
                f'{self.url}/{template}/install.json',
                destination='install.json',
                overwrite='prompt',
                default_choice='no',
            )
            self.download_file(
                f'{self.url}/{template}/tcex.json',
                destination='tcex.json',
                overwrite='prompt',
                default_choice='no',
            )

    def init_external(self, template: str):
        """Download external App files."""
        self.download_file(
            f'{self.url}/external/external_app.py', destination='external_app.py', overwrite=True,
        )
        self.download_file(f'{self.url}/external/run.py', destination='run.py', overwrite=True)
        self.download_file(
            f'{self.url}/{template}/app_config.json',
            destination='app_config.json',
            overwrite=False,
        )

    def init_job(self):
        """Download job App files."""
        self.download_file(f'{self.url}/__main__.py', destination='__main__.py', overwrite=True)
        self.download_file(f'{self.url}/job/job_app.py', destination='job_app.py', overwrite=True)
        self.download_file(f'{self.url}/job/run.py', destination='run.py', overwrite=True)

    def init_playbook(self):
        """Download playbook App files."""
        self.download_file(f'{self.url}/__main__.py', destination='__main__.py', overwrite=True)
        self.download_file(
            f'{self.url}/playbook/playbook_app.py', destination='playbook_app.py', overwrite=True,
        )
        self.download_file(f'{self.url}/playbook/run.py', destination='run.py', overwrite=True)

    def init_service(self, template: str):
        """Download service App files."""
        self.download_file(
            f'{self.url}/service/service_app.py', destination='service_app.py', overwrite=True
        )
        self.download_file(f'{self.url}/{template}/run.py', destination='run.py', overwrite=True)

    def test_conftest_py(self):
        """Download the configtest.py file."""
        if not self.profile:
            raise RuntimeError('Download of "conftest.py" requires a Profile object.')
        url = f'{self.url}/tests/conftest.py'
        destination: str = os.path.join(self.profile.test_directory, 'conftest.py')
        self.download_file(url, destination, overwrite=True)

    def test_profiles_py(self):
        """Download the profiles.py file."""


class TestProfileTemplates(TemplateBase):
    """TestProfile Template Class

    Args:
        branch: The git branch of tcex to use for downloads.
        profile: An instance of Profile.
    """

    def __init__(self, profile: Optional[object] = None, branch: Optional[str] = 'master'):
        """Initialize Class properties."""
        super().__init__(profile, branch)

    @property
    def app_class(self) -> str:
        """Return the proper App class based on runtime level."""
        app_type_to_class = {
            'organization': 'TestCaseJob',
            'playbook': 'TestCasePlaybook',
            'triggerservice': 'TestCaseTriggerService',
            'webhooktriggerservice': 'TestCaseWebhookTriggerService',
        }
        return app_type_to_class.get(self.profile.ij.runtime_level.lower())

    def render_templates(self):
        """Render the templates and write to disk conditionally."""
        self.test_profiles_py()

    def test_profiles_py(self):
        """Render the test_profiles.py template."""
        filename: str = os.path.join(self.profile.feature_directory, 'test_profiles.py')
        variables = {
            'class_name': self.app_class,
            'runtime_level': self.profile.ij.runtime_level.lower(),
        }
        self.render(f'{self.url}/tests/test_profiles.py.tpl', filename, variables, True)


class ValidationTemplates(TemplateBase):
    """Validation Template Class

    Args:
        profile (Profile): An instance of the Profile object.
        branch (str, optional): The tcex git branch. Defaults to 'master'.

    Raises:
        NotImplementedError: The delete method is not currently implemented.
    """

    def __init__(self, profile: Optional[object] = None, branch: Optional[str] = 'master'):
        """Initialize Class properties."""
        super().__init__(profile, branch)

        # properties
        self.utils = Utils()

    def output_data(self, output_variables: Optional[list]):
        """Return formatted output data.

        variable format: #App:9876:http.content!Binary
        """
        output_data = []
        for ov in output_variables:
            output_data.append({'method': self.utils.variable_method_name(ov), 'variable': ov})
        return sorted(output_data, key=lambda i: i['method'])

    def validate_py(self):
        """Render the validate.py template."""
        filename: str = os.path.join(self.profile.test_directory, 'validate.py')
        variables = {
            'feature': self.profile.feature,
            'output_data': self.output_data(self.profile.ij.tc_playbook_out_variables),
        }
        self.render(f'{self.url}/tests/validate.py.tpl', filename, variables, True)

    def validate_custom_py(self):
        """Render the validate_custom.py template."""
        filename: str = os.path.join(self.profile.test_directory, 'validate_custom.py')
        variables = {
            'feature': self.profile.feature,
            'output_data': self.output_data(self.profile.ij.tc_playbook_out_variables),
        }
        self.render(f'{self.url}/tests/validate_custom.py.tpl', filename, variables)

    def validate_feature_py(self):
        """Render the validate_custom.py template."""
        filename: str = os.path.join(self.profile.feature_directory, 'validate_feature.py')
        variables = {
            'feature': self.profile.feature,
            'output_data': self.output_data(self.profile.ij.tc_playbook_out_variables),
        }
        self.render(f'{self.url}/tests/validate_feature.py.tpl', filename, variables)

    def render_templates(self):
        """Render the templates and write to disk conditionally."""
        self.validate_py()
        self.validate_custom_py()
        self.validate_feature_py()



#!/usr/bin/env python
"""TcEx Framework Test Generation Module."""
# standard library
import copy
import json
import os

from ..app_config_object.templates import (
    CustomTemplates,
    DownloadTemplates,
    TestProfileTemplates,
    ValidationTemplates,
)
from ..profile import Interactive, Profile
from .bin import Bin


class Test(Bin):
    """Create testing files for ThreatConnect Exchange App.

    Args:
        _args (namespace): The argparser args Namespace.
    """

    def __init__(self, _args):
        """Initialize Class properties."""
        super().__init__(_args)

        # properties
        self.profile = Profile(
            default_args={}, feature=self.args.feature, name=self.args.profile_name
        )
        self.custom_templates = CustomTemplates(self.profile, self.args.branch)
        self.download_template = DownloadTemplates(self.profile, self.args.branch)
        self.profile_interactive = Interactive(self.profile)
        self.test_profile_template = TestProfileTemplates(self.profile, self.args.branch)
        self.validation_templates = ValidationTemplates(self.profile, self.args.branch)

    def add_negative_profile(self, profile_name, inputs, fail_on_error=None):
        """Create a negative profile."""
        # build profile name
        exit_code = 1  # default exit code is 1
        if fail_on_error is not None:
            if fail_on_error is False:
                exit_code = 0
            profile_name = f'{profile_name}_foe_{str(fail_on_error).lower()}'

        # get profile data and update with new inputs
        profile_data = self.profile.contents
        profile_data['exit_codes'] = [exit_code]
        profile_data['exit_message'] = None
        profile_data['inputs'] = inputs
        profile_data['outputs'] = None

        # create a meaningful profile name
        new_profile = Profile(default_args={}, feature=self.args.feature, name=profile_name)
        new_profile.add(profile_data=profile_data)

    def create_dirs(self):
        """Create tcex.d directory and sub directories."""
        for d in [
            self.profile.test_directory,
            self.profile.feature_directory,
            self.profile.directory,
        ]:
            if not os.path.isdir(d):
                os.makedirs(d)

        # create __init__ files
        self.create_dirs_init()

    def create_dirs_init(self):
        """Create the __init__.py file under dir."""
        for d in [self.profile.test_directory, self.profile.feature_directory]:
            if os.path.isdir(d):
                with open(os.path.join(d, '__init__.py'), 'a'):
                    os.utime(os.path.join(d, '__init__.py'), None)

    def create_negative_profiles(self):
        """Create negative profiles using interactive profile base."""
        for inputs in self.profile.profile_inputs:
            for name, value in inputs.get('required', {}).items():
                ij_data = self.profile.ij.params_dict.get(name, {})
                # create a profile for each pb data type
                for pb_data_type in ij_data.get('playbookDataType', []):
                    for negative_type in self.negative_inputs.get(pb_data_type.lower(), []):
                        # the value is pre-staged in test_case_playbook_common.py
                        value = f'#App:1234:{negative_type}!{pb_data_type}'
                        profile_name = f'negative_{name}_{pb_data_type.lower()}_{negative_type}'
                        # modify copy so original is preserved for next interation
                        new_inputs = copy.deepcopy(inputs)
                        new_inputs['required'][name] = value

                        if 'fail_on_error' in inputs.get('optional', {}):
                            # handle fail on error
                            for b in [False, True]:
                                new_inputs['optional']['fail_on_error'] = b
                                self.add_negative_profile(profile_name, new_inputs, b)
                        else:
                            self.add_negative_profile(profile_name, new_inputs)

    def interactive_profile(self, negative=False):
        """Present interactive profile inputs."""
        self.profile_interactive.present()
        profile_data = {
            'exit_codes': self.profile_interactive.exit_codes,
            'inputs': self.profile_interactive.inputs,
            'stage': self.profile_interactive.staging_data,
        }
        self.profile.add(profile_data=profile_data)

        if negative:
            # if user specified negative arg then create negative test profiles
            self.create_negative_profiles()

    @staticmethod
    def load_legacy_profiles(staging_files):
        """Load staging data to migrate legacy templates."""
        staging_data = {}
        for sf in staging_files:
            with open(sf) as fh:
                data = json.load(fh)
            for d in data:
                staging_data[d.get('variable')] = d.get('data')
        return staging_data

    def migrate_profile(self):
        """Migrate legacy profile to new framework."""
        data = []
        profile_file = os.path.join(self.app_path, 'tcex.d', 'profiles', self.args.profile_file)
        if os.path.isfile(self.args.profile_file):
            with open(self.args.profile_file) as fh:
                data = json.load(fh)
        elif os.path.isfile(profile_file):
            with open(profile_file) as fh:
                data = json.load(fh)
        else:
            self.handle_error(f'Error reading in profile file: {self.args.profile_file}', True)

        for d in data:
            profile_data = {
                'exit_codes': d.get('exit_codes'),
                'exit_message': None,
                'inputs': d.get('args', {}).get('app'),
                'stage': {'kvstore': self.load_legacy_profiles(d.get('data_files', []))},
            }

            # add profile
            self.profile.add(profile_data=profile_data, profile_name=d.get('profile_name'))

    @property
    def negative_inputs(self):
        """Return dict of negative inputs."""
        return {
            'binary': ['empty', 'null'],
            'binaryarray': ['empty', 'null'],
            'keyvalue': ['null'],
            'keyvaluearray': ['null'],
            'string': ['empty', 'null'],
            'stringarray': ['empty', 'null'],
            'tcentity': ['null'],
            'tcentityarray': ['null'],
        }



"""Declares several transform functions that verify an argument is of a specified type."""
# standard library
from typing import Any

# first-party
from tcex.utils.utils import Utils

from .validation_exception import ValidationError


def _to_type(target_type, **kwargs):
    """Validate that an argument can be cast to a type by calling type(value).

    Accepts either a String or StringArray value.  If A StringArray, validates that every element
    in the array can be cast to the given type.

    Args:
        target_type (class):  type to invoke
        allow_none (bool): If none values are ok. default: False

    Returns:
        A transform function that can be used in the transforms argument to @ReadArg.
    """

    def _transform(value: Any, arg_name: str, label: str):
        """Run transform on input data.

        Args:
            value: The input data to be transformed.
            arg_name: The name of the input arg.
            label: The label displayed to the user.

        Raises:
            ValidationError: raised on validation failure.
        """
        allow_none = kwargs.get('allow_none', False)

        if not isinstance(value, list):
            if (value is None or value == '') and allow_none:
                return value

            try:
                return target_type(value)
            except Exception:
                raise ValidationError(f'"{label}" ({arg_name}) must be a {target_type.__name__}.')

        transformed = []
        for v in value:
            if (v is None or v == '') and allow_none:
                transformed.append(v)
                continue

            try:
                transformed.append(target_type(v))
            except Exception:
                raise ValidationError(f'"{label}" ({arg_name}) must be a {target_type.__name__}.')

        return transformed

    return _transform


def to_int(allow_none=False):
    """Transform an argument into an int

    Allowed argument types: String, StringArray

    Args:
        allow_none (bool): skip any None elements

    Returns:
        A transform function that can be used in the transforms argument to @ReadArg.
    """
    return _to_type(int, allow_none=allow_none)


def to_float(allow_none=False):
    """Transform an argument value into a float.

    Allowed argument types: String, StringArray

    Args:
        allow_none (bool): skip any None elements

    Returns:
        A transform function that can be used in the transforms argument to @ReadArg.
    """
    return _to_type(float, allow_none=allow_none)


def to_bool():
    """Transform an argument value into a bool.

    Allowed argument types: String, StringArray

    Returns:
        A transform function that can be used in the transforms argument to @ReadArg.
    """

    def _validator(value, arg_name, label):  # pylint: disable=unused-argument
        if not isinstance(value, list):
            return Utils.to_bool(value)

        return list([Utils.to_bool(v) for v in value])

    return _validator



"""TcEx Playbook Test Case module"""
# standard library
import os
import subprocess
import sys

from .test_case_playbook_common import TestCasePlaybookCommon


class TestCasePlaybook(TestCasePlaybookCommon):
    """Playbook TestCase Class"""

    run_method = 'inline'  # run service inline or a as subprocess

    def run(self):
        """Run the Playbook App.

        Args:
            args (dict): The App CLI args.

        Returns:
            int: The exit code fo the App.
        """
        # first-party
        from run import run  # pylint: disable=no-name-in-module

        # run the app
        exit_code = 0
        try:
            run()
        except SystemExit as e:
            exit_code = e.code

        self.log.data('run', 'Exit Code', exit_code)
        return exit_code

    def run_profile(self):
        """Run an App using the profile name."""
        # create encrypted config file
        self.create_config(self.profile.args)

        # run the service App in 1 of 3 ways
        if self.run_method == 'inline':
            # backup sys.argv
            sys_argv_orig = sys.argv

            # clear sys.argv
            sys.argv = sys.argv[:1]

            # run the App
            exit_code = self.run()

            # restore sys.argv
            sys.argv = sys_argv_orig
        elif self.run_method == 'subprocess':
            # run the Service App as a subprocess
            app_process = subprocess.Popen(['python', 'run.py'])
            exit_code = app_process.wait()

        try:
            # remove env var for encrypted file if there
            del os.environ['TC_APP_PARAM_KEY']
            del os.environ['TC_APP_PARAM_FILE']
        except KeyError:
            pass

        # add context for populating output variables
        self.profile.add_context(self.context)

        return exit_code

    def setup_method(self):
        """Run before each test method runs."""
        super().setup_method()
        self.stager.redis.from_dict(self.redis_staging_data)

        self.redis_client = self.tcex.redis_client



"""ThreatConnect TI Indicator"""
# standard library
import json
from urllib.parse import quote, unquote

from ..mappings import Mappings

# import local modules for dynamic reference
module = __import__(__name__)


def custom_indicator_class_factory(
    indicator_type, entity_type, branch_type, base_class, value_fields
):
    """Build dynamic Custom Indicator Class."""

    @staticmethod
    def _metadata_map_1():
        """Map field data."""
        metadata_map = base_class._metadata_map()
        for value in value_fields:
            manipulated_value = value.lower().replace(' ', '_')
            if manipulated_value not in metadata_map.keys():
                metadata_map[manipulated_value] = value
        return metadata_map

    def init(self, tcex, **kwargs):  # pylint: disable=possibly-unused-variable
        """Init method for Custom Indicator Types with one value"""
        base_class.__init__(
            self,
            tcex,
            sub_type=indicator_type,
            api_entity=entity_type,
            api_branch=branch_type,
            **kwargs,
        )
        res = {v: k for k, v in self._metadata_map().items()}
        values = []
        for field in value_fields:
            value = kwargs.pop(res.get(field), kwargs.pop(field, ''))
            value = quote(self.fully_decode_uri(value), safe='')
            values.append(value)

        if len(values) == 1:
            self.unique_id = kwargs.get('unique_id', values[0])
        elif len(values) == 2:
            self.unique_id = kwargs.get('unique_id', self.build_summary(values[0], values[1]))
        elif len(values) == 3:
            self.unique_id = kwargs.get(
                'unique_id', self.build_summary(values[0], values[1], values[2])
            )

    def _set_unique_id(self, json_request):
        """Set the unique ID.

        Args:
            json_request (dict): The JSON data for the request.
        """
        values = []
        for field in value_fields:
            value = json_request.get(field, '')
            values.append(quote(self.fully_decode_uri(value), safe=''))
        if len(values) == 1:
            self.unique_id = values[0]
        elif len(values) == 2:
            self.unique_id = self.build_summary(values[0], values[1])
        elif len(values) == 1:
            self.unique_id = self.build_summary(values[0], values[1], values[2])

    def can_create(self):  # pylint: disable=unused-argument,possibly-unused-variable
        """Determine if the required data that the API endpoint is expecting is present."""
        valid_create = True
        for field in value_fields:
            if not field:
                valid_create = False
        return valid_create

    class_name = indicator_type.replace(' ', '')
    init_method = locals()['init']
    set_unique_id_method = locals()['_set_unique_id']
    can_create_method = locals()['can_create']
    _metadata_map = locals()['_metadata_map_1']
    new_class = type(
        str(class_name),
        (base_class,),
        {
            '__init__': init_method,
            '_set_unique_id': set_unique_id_method,
            'can_create': can_create_method,
            '_metadata_map': _metadata_map,
        },
    )
    return new_class


class Indicator(Mappings):
    """Unique API calls for Indicator API Endpoints"""

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties."""
        super().__init__(
            tcex,
            'Indicator',
            'indicators',
            kwargs.pop('sub_type', None),
            kwargs.pop('api_entity', 'indicator'),
            kwargs.pop('api_branch', None),
            kwargs.pop('owner', None),
        )

        for arg, value in kwargs.items():
            self.add_key_value(arg, value)

    @property
    def as_entity(self):
        """Return the entity representation of the Indicator."""
        return {
            'type': self.api_sub_type,
            'value': unquote(self.unique_id),
            'id': self._data.get('id'),
        }

    @staticmethod
    def is_indicator():
        """Return true if object type is an indicator."""
        return True

    @property
    def owner(self):
        """Return the owner."""
        return self._owner

    def can_create(self):
        """Overridden by other indicator classes."""
        return True

    @staticmethod
    def _metadata_map():
        """Map snake case fields to camel case fields."""
        return {
            'date_added': 'dateAdded',
            'dns_active': 'dnsActive',
            'last_modified': 'lastModified',
            'private_flag': 'privateFlag',
            'whois_active': 'whoisActive',
            'key_name': 'Key Name',
            'value_type': 'Value Type',
            'value_name': 'Value Name',
            'block': 'Block',
            'mutex': 'Mutex',
            'as_number': 'AS Number',
            'hostname': 'hostName',
        }

    def add_key_value(self, key, value):
        """Convert the value and adds it as a data field.

        Args:
            key:
            value:
        """
        key = self._metadata_map().get(key, key)
        if key in ['dateAdded', 'lastModified']:
            self._data[key] = self._utils.datetime.format_datetime(
                value, date_format='%Y-%m-%dT%H:%M:%SZ'
            )
        elif key == 'confidence':
            self._data[key] = int(value)
        elif key == 'rating':
            self._data[key] = float(value)
        elif key == 'unique_id':
            self._unique_id = quote(self.fully_decode_uri(value), safe='')
        else:
            self._data[key] = value

    def status(self, status=None, cal_status=None):
        """Update the Indicators status

        Args:
            status:  Valid values to set to active are ['active', '2', '1' ] while
            ['inactive', '-2', '-1', 0] will set it to inactive
            cal_status:  Valid values to set to locked are ['locked', 'lock', '1' ] while
            ['unlock', 'unlocked', '0'] will set it to inactive

        Returns:
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])
        if not status and not cal_status:
            return None
        request_data = {}
        if status:
            status = str(status)
            if status.lower() in ['active', '1']:
                request_data['active'] = 1
            elif status.lower() in ['inactive', '0']:
                request_data['active'] = 0
        if cal_status:
            cal_status = str(cal_status)
            if cal_status.lower() in ['locked', 'lock', '1']:
                request_data['activeLocked'] = 1
            elif cal_status.lower() in ['unlock', 'unlocked', '0']:
                request_data['activeLocked'] = 0
        return self.tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, request_data, owner=self.owner
        )

    def rating(self, value):
        """Update the Indicators rating

        Args:
            value:
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])
        request_data = {'rating': value}
        return self.tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, request_data, owner=self.owner
        )

    def confidence(self, value):
        """Update the Indicators confidence

        Args:
            value:
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])
        request_data = {'confidence': value}
        return self.tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, request_data, owner=self.owner
        )

    def owners(self):
        """Return owners"""
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])
        return self.tc_requests.owners(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner
        )

    def add_observers(self, count, date_observed):
        """Add a Indicator Observation

        Args:
            count:
            date_observed:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        data = {
            'count': count,
            'dateObserved': self._utils.datetime.format_datetime(
                date_observed, date_format='%Y-%m-%dT%H:%M:%SZ'
            ),
        }

        return self.tc_requests.add_observations(
            self.api_type, self.api_branch, self.unique_id, data, owner=self.owner
        )

    def observation_count(self):
        """Get the indicators observation count.

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])
        return self.tc_requests.observation_count(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner
        )

    def add_false_positive(self):
        """Add a Indicator FalsePositive."""
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.add_false_positive(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner
        )

    def observations(self):
        """Return indicator observation data.

        Returns:
            [type]: [description]
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])
        return self.tc_requests.observations(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner
        )

    def deleted(self, deleted_since=None, filters=None, params=None):
        """Return deleted indicators from TC REST API.

        Args:
            deleted_since ([type]): [description]
            filters ([type], optional): [description]. Defaults to None.
            params ([type], optional): [description]. Defaults to None.

        Returns:
            [type]: [description]
        """
        return self.tc_requests.deleted(
            self.api_type,
            self.api_branch,
            deleted_since=deleted_since,
            owner=self.owner,
            filters=filters,
            params=params,
        )

    @staticmethod
    def build_summary(val1=None, val2=None, val3=None):
        """Construct an indicator summary given va1, va2, val3.

        Args:
            val1 (str, optional): Indicator value. Defaults to None.
            val2 (str, optional): Indicator value. Defaults to None.
            val3 (str, optional): Indicator value. Defaults to None.

        Returns:
            str: <space><colon><space> delimeted indicator summary.
        """
        summary = []
        if val1 is not None:
            summary.append(val1)
        if val2 is not None:
            summary.append(val2)
        if val3 is not None:
            summary.append(val3)
        if not summary:
            return None
        return ' : '.join(summary)

    def __str__(self):
        """Return string representation of object"""
        return json.dumps(self._data, indent=4)



"""App Config Object module for TcEx Framework"""
# flake8: noqa
from .install_json import InstallJson
from .layout_json import LayoutJson
from .permutations import Permutations
from .tcex_json import TcexJson



"""ThreatConnect Case Management Filter Module"""


class Filter:
    """Filter Base Class

    Args:
        tql (TQL): Instance of TQL Class.
    """

    def __init__(self, api_endpoint, tcex, tql):
        """Initialize Class properties"""
        self._api_endpoint = api_endpoint.value
        self._tcex = tcex
        self._tql = tql
        self._tql_data = None
        self.artifacts = None

    @property
    def implemented_keywords(self):
        """Return implemented TQL keywords."""
        keywords = []
        for prop in dir(self):
            if prop.startswith('_') or prop in ['tql']:
                continue
            keywords.append(prop)

        return keywords

    @property
    def keywords(self):
        """Return supported TQL keywords."""
        return [td.get('keyword') for td in self.tql_data]

    @property
    def tql_data(self):
        """Return TQL data keywords."""
        if self._tql_data is None:
            r = self._tcex.session.options(f'{self._api_endpoint}/tql', params={})
            if r.ok:
                self._tql_data = r.json()['data']

        return self._tql_data

    def tql(self, tql):
        """Filter objects based on TQL expression.

        Args:
            tql (str): The raw TQL string for the filter.
        """
        self._tql.set_raw_tql(tql)
